/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// common.c -- misc functions used in client and server

#include "quakedef.h"

#define NUM_SAFE_ARGVS  7

static char     *largv[MAX_NUM_ARGVS + NUM_SAFE_ARGVS + 1];
static char     *argvdummy = " ";
static char     *safeargvs[NUM_SAFE_ARGVS] = {
		"-stdvid", 
		"-nolan", 
		"-nosound", 
		"-nocdaudio", 
		"-nojoy", 
		"-nomouse", 
		"-dibonly"
};

cvar_t  registered = {"registered","0"};
cvar_t  cmdline = {"cmdline","0", false, true};

// if a packfile directory differs from this, it is assumed to be hacked
#define PAK0_COUNT              339
#define PAK0_CRC                32981

char	com_token[1024];
int		com_argc;
char	**com_argv;

#define CMDLINE_LENGTH	256
char		com_cmdline[CMDLINE_LENGTH];

qboolean    bigendian;
qboolean	standard_quake = true, rogue = false, hipnotic = false;

void COM_Path_f (void);
void COM_Gamedir_f (void);

/*

All of Quake's data access is through a hierchal file system, but the contents of the file system can be transparently merged from several sources.

The "base directory" is the path to the directory holding the quake.exe and all game directories.  The sys_* files pass this to host_init in quakeparms_t->basedir.  This can be overridden with the "-basedir" command line parm to allow code debugging in a different directory.  The base directory is
only used during filesystem initialization.

The "game directory" is the first tree on the search path and directory that all generated files (savegames, screenshots, demos, config files) will be saved to.  This can be overridden with the "-game" command line parameter.  The game directory can never be changed while quake is executing.  This is a precacution against having a malicious server instruct clients to write files over areas they shouldn't.

The "cache directory" is only used during development to save network bandwidth, especially over ISDN / T1 lines.  If there is a cache directory
specified, when a file is found by the normal search path, it will be mirrored
into the cache directory, then opened there.

FIXME:
The file "parms.txt" will be read out of the game directory and appended to the current command line arguments to allow different games to initialize startup parms differently.  This could be used to add a "-sspeed 22050" for the high quality sound edition.  Because they are added at the end, they will not override an explicit setting on the original command line.

*/

//============================================================================


// ClearLink is used for new headnodes
void ClearLink (link_t *l)
{
	l->prev = l->next = l;
}

void RemoveLink (link_t *l)
{
	l->next->prev = l->prev;
	l->prev->next = l->next;
}

void InsertLinkBefore (link_t *l, link_t *before)
{
	l->next = before;
	l->prev = before->prev;
	l->prev->next = l;
	l->next->prev = l;
}
void InsertLinkAfter (link_t *l, link_t *after)
{
	l->next = after->next;
	l->prev = after;
	l->prev->next = l;
	l->next->prev = l;
}

/*
============================================================================

					LIBRARY REPLACEMENT FUNCTIONS

============================================================================
*/

// fast safe printbuffers courtesy of lord havoc.
int va_snprintf (char *function, char *buffer, size_t buffersize, const char *format, ...)
{
	va_list args;
	int		result;
	
	va_start (args, format);
	result = va_vsnprintf (function, buffer, buffersize, format, args);
	va_end (args);
	
	return result;
}

int va_vsnprintf (char *function, char *buffer, size_t buffersize, const char *format, va_list args)
{
	size_t result;
	
	result = vsnprintf (buffer, buffersize, format, args);
	
	if (result < 0 || result >= buffersize)
	{
		static qboolean inside;
		
		// Beware recursion here
		if (!inside)
		{
			inside = true;
			Con_SafePrintf ("%s: excessive string length, max = %d\n", function, buffersize);
		}		
		inside = false;
		buffer[buffersize - 1] = '\0';
		return -1;
	}	
	return result;
}

// be very carefull especially in host.c cmd.c and cvar.c
#ifndef HAVE_STRLCAT
size_t strlcat(char *dst, const char *src, size_t siz)
{
	register char		*d = dst;
	register const char *s = src;
	register size_t		n = siz;
	size_t				dlen;

	/* Find the end of dst and adjust bytes left but don't go past end */
	while (n-- != 0 && *d != '\0')
	{
		d++;
	}
	dlen = d - dst;
	n = siz - dlen;

	if (n == 0)
	{
		return(dlen + strlen(s));
	}

	while (*s != '\0') 
	{
		if (n != 1) 
		{
			*d++ = *s;
			n--;
		}
		s++;
	}
	*d = '\0';

	return(dlen + (s - src));	/* count does not include NUL */
}
#endif  // #ifndef HAVE_STRLCAT

// be very carefull especially in host.c cmd.c and cvar.c
#ifndef HAVE_STRLCPY
size_t strlcpy(char *dst, const char *src, size_t siz)
{
	register char		*d = dst;
	register const char *s = src;
	register size_t		n = siz;

	/* Copy as many bytes as will fit */
	if (n != 0 && --n != 0) 
	{
		do 
		{
			if ((*d++ = *s++) == 0)	break;
		} while (--n != 0);
	}

	/* Not enough room in dst, add NUL and traverse rest of src */
	if (n == 0) 
	{
		if (siz != 0)
		{
			*d = '\0';		/* NUL-terminate dst */
		}
		while (*s++);
	}	
	return(s - src - 1);	/* count does not include NUL */
}
#endif  // #ifndef HAVE_STRLCPY

/*
============================================================================

					BYTE ORDER FUNCTIONS

============================================================================
*/
short   (*BigShort) (short l);
short   (*LittleShort) (short l);
int     (*BigLong) (int l);
int     (*LittleLong) (int l);
float   (*BigFloat) (float l);
float   (*LittleFloat) (float l);

short ShortSwap (short l)
{
	byte    b1,b2;

	b1 = l&255;
	b2 = (l>>8)&255;

	return (b1<<8) + b2;
}

short ShortNoSwap (short l)
{
	return l;
}

int LongSwap (int l)
{
	byte    b1,b2,b3,b4;

	b1 = l&255;
	b2 = (l>>8)&255;
	b3 = (l>>16)&255;
	b4 = (l>>24)&255;

	return ((int)b1<<24) + ((int)b2<<16) + ((int)b3<<8) + b4;
}

int	LongNoSwap (int l)
{
	return l;
}

float FloatSwap (float f)
{
	union
	{
		float   f;
		byte    b[4];
	} dat1, dat2;
	
	
	dat1.f = f;
	dat2.b[0] = dat1.b[3];
	dat2.b[1] = dat1.b[2];
	dat2.b[2] = dat1.b[1];
	dat2.b[3] = dat1.b[0];
	return dat2.f;
}

float FloatNoSwap (float f)
{
	return f;
}

/*
==============================================================================

			MESSAGE IO FUNCTIONS

Handles byte ordering and avoids alignment errors
==============================================================================
*/

//
// writing functions
//

void MSG_WriteChar (sizebuf_t *sb, int c)
{
	byte    *buf;
	buf = SZ_GetSpace (sb, 1);
	buf[0] = c;
}

void MSG_WriteByte (sizebuf_t *sb, int c)
{
	byte    *buf;
	buf = SZ_GetSpace (sb, 1);
	buf[0] = c;
}

void MSG_WriteShort (sizebuf_t *sb, int c)
{
	byte    *buf;
	buf = SZ_GetSpace (sb, 2);
	buf[0] = c&0xff;
	buf[1] = c>>8;
}

void MSG_WriteByteShort (sizebuf_t *sb, int c)
{
	int PrevSize = sv.signon.cursize;
	
	if (sv.protocol == PROTOCOL_VERSION)
	{
		MSG_WriteByte (sb, c);
	}
	else
	{
		MSG_WriteShort (sb, c); // PROTOCOL_VERSION_EXT-3
	}

	if (sb == &sv.signon)
	{
		sv.signondiff += sv.signon.cursize - (PrevSize + 1); // Track extra bytes due to >256 model support, kludge
	}
}


void MSG_WriteByteShort2 (sizebuf_t *sb, int c, qboolean large)
{
	if (sv.protocol < PROTOCOL_VERSION_EXT2 || large && sv.protocol > PROTOCOL_VERSION_EXT2)
	{
		MSG_WriteByte (sb, c); // Some progs (Marcher) send sound services, maintain compatibility, kludge
	}
	else
	{
		MSG_WriteShort (sb, c);
	}
}

void MSG_WriteLong (sizebuf_t *sb, int c)
{
	byte    *buf;
	
	buf = SZ_GetSpace (sb, 4);
	buf[0] = c&0xff;
	buf[1] = (c>>8)&0xff;
	buf[2] = (c>>16)&0xff;
	buf[3] = c>>24;
}

void MSG_WriteFloat (sizebuf_t *sb, float f)
{
	union
	{
		float   f;
		int     l;
	} dat;
	
	
	dat.f = f;
	dat.l = LittleLong (dat.l);
	
	SZ_Write (sb, &dat.l, 4);
}

void MSG_WriteString (sizebuf_t *sb, char *s)
{
	if (!s)
	{
		SZ_Write (sb, "", 1);
	}
	else
	{
		SZ_Write (sb, s, strlen(s)+1);
	}
}

void MSG_WriteCoord (sizebuf_t *sb, float f)
{
	MSG_WriteShort (sb, (int)(f*8));
}

void MSG_WriteAngle (sizebuf_t *sb, float f)
{
	MSG_WriteByte (sb, ((int)f*256/360) & 255);
}

//
// reading functions
//
int             msg_readcount;
qboolean        msg_badread;

void MSG_BeginReading (void)
{
	msg_readcount = 0;
	msg_badread = false;
}

// returns -1 and sets msg_badread if no more characters are available
int MSG_ReadChar (void)
{
	int     c;
	
	if (msg_readcount+1 > net_message.cursize)
	{
		msg_badread = true;
		return -1;
	}		
	c = (signed char)net_message.data[msg_readcount];
	msg_readcount++;
	
	return c;
}


// like msg_readbyte but does not advance the counter
int MSG_SnoopByte (void)
{
	if (msg_readcount + 1 > net_message.cursize)
	{
		return -1;
	}
	return (unsigned char) net_message.data[msg_readcount];
}


int MSG_ReadByte (void)
{
	int     c;
	
	if (msg_readcount+1 > net_message.cursize)
	{
		msg_badread = true;
		return -1;
	}		
	c = (unsigned char)net_message.data[msg_readcount];
	msg_readcount++;
	
	return c;
}

int MSG_ReadByteShort (void)
{
	return cl.protocol == PROTOCOL_VERSION ? MSG_ReadByte () : MSG_ReadShort ();
}

int MSG_ReadByteShort2 (qboolean large)
{
	if (cl.protocol < PROTOCOL_VERSION_EXT2 || large && cl.protocol > PROTOCOL_VERSION_EXT2)
	{
		return MSG_ReadByte (); // Some progs (Marcher) send sound services, maintain compatibility, kludge
	}
	else
	{
		return MSG_ReadShort ();
	}
}

int MSG_ReadShort (void)
{
	int     c;
	
	if (msg_readcount+2 > net_message.cursize)
	{
		msg_badread = true;
		return -1;
	}		
	c = (short)(net_message.data[msg_readcount] + (net_message.data[msg_readcount+1] << 8));
	msg_readcount += 2;
	
	return c;
}

int MSG_ReadLong (void)
{
	int     c;
	
	if (msg_readcount+4 > net_message.cursize)
	{
		msg_badread = true;
		return -1;
	}		
	c = net_message.data[msg_readcount]
	+ (net_message.data[msg_readcount+1]<<8)
	+ (net_message.data[msg_readcount+2]<<16)
	+ (net_message.data[msg_readcount+3]<<24);
	
	msg_readcount += 4;
	
	return c;
}

float MSG_ReadFloat (void)
{
	union
	{
		byte    b[4];
		float   f;
		int     l;
	} dat;
	
	dat.b[0] =      net_message.data[msg_readcount];
	dat.b[1] =      net_message.data[msg_readcount+1];
	dat.b[2] =      net_message.data[msg_readcount+2];
	dat.b[3] =      net_message.data[msg_readcount+3];
	msg_readcount += 4;
	
	dat.l = LittleLong (dat.l);

	return dat.f;   
}

char *MSG_ReadString (void)
{
	static char     string[2048];
	int             l,c;
	
	l = 0;

	do
	{
		c = MSG_ReadChar ();
		if (c == -1 || c == 0) break;
		string[l] = c;
		l++;
	} while (l < sizeof(string)-1);
	
	string[l] = 0;
	
	return string;
}

float MSG_ReadCoord (void)
{
	return MSG_ReadShort() * (1.0/8);
}

float MSG_ReadAngle (void)
{
	return MSG_ReadChar() * (360.0/256);
}



//===========================================================================

void SZ_Alloc (sizebuf_t *buf, int startsize)
{
	if (startsize < 256)
	{
		startsize = 256;
	}
	buf->data = Heap_MemAlloc(startsize);
	memset (buf->data, 0, startsize);
	buf->maxsize = startsize;
	buf->cursize = 0;
}

void SZ_Clear (sizebuf_t *buf)
{
	buf->cursize = 0;
}

void *SZ_GetSpace (sizebuf_t *buf, int length)
{
	void    *data;
	
	if (buf->cursize + length > buf->maxsize)
	{
		if (!buf->allowoverflow)
		{
			Sys_Error ("SZ_GetSpace: overflow without allowoverflow set");
		}

		if (length > buf->maxsize)
		{
			Sys_Error ("SZ_GetSpace: %i is > full buffer size", length);
		}
		buf->overflowed = true;
		Con_Printf ("SZ_GetSpace: overflow");
		SZ_Clear (buf); 
	}
	data = buf->data + buf->cursize;
	buf->cursize += length;
	
	return data;
}

void SZ_Write (sizebuf_t *buf, void *data, int length)
{
	memcpy (SZ_GetSpace(buf, length), data, length);         
}

void SZ_Print (sizebuf_t *buf, char *data)
{
	int  len;
	
	len = strlen(data) + 1;

	// byte * cast to keep VC++ happy
	if (buf->data[buf->cursize-1])
	{
		memcpy ((byte *)SZ_GetSpace(buf, len), data, len); // no trailing 0
	}
	else
	{
		memcpy ((byte *)SZ_GetSpace(buf, len - 1) - 1, data, len); // write over trailing 0
	}
}


//============================================================================

/*
============
COM_Timeout
============
*/
qboolean COM_Timeout (float *PrevTime, float WaitTime)
{
	float CurrTime;
	
	CurrTime = Sys_FloatTime ();
	
	if (*PrevTime && CurrTime - *PrevTime < WaitTime)
	{
		return false;
	}
	*PrevTime = CurrTime;
	
	return true;
}

/*
============
COM_SkipPath
============
*/
char *COM_SkipPath (char *pathname)
{
	char    *last;
	
	last = pathname;

	while (*pathname)
	{
		if (*pathname=='/')
		{
			last = pathname+1;
		}
		pathname++;
	}
	return last;
}

/*
============
COM_StripExtension
============
*/
void COM_StripExtension (char *in, char *out)
{
	while (*in && *in != '.')
	{
		*out++ = *in++;
	}
	*out = '\0';
}

/*
============
COM_FileExtension
============
*/
char *COM_FileExtension (char *in)
{
	static char		exten[8];
	int             i;

	while (*in && *in != '.')
	{
		in++;
	}

	if (!*in)
	{
		return "";
	}
	in++;

	for (i=0 ; i<7 && *in ; i++,in++)
	{
		exten[i] = *in;
	}
	exten[i] = '\0';

	return exten;
}

/*
============
COM_FileBase
============
*/
void COM_FileBase (const char *in, char *out)
{
	const char *start, *end;
	int	length;

	if (!(end = strrchr(in, '.')))
	{
		end = in + strlen(in);
	}

	if (!(start = strrchr(in, '/')))
	{
		start = in;
	}
	else
	{
		start += 1;
	}
	length = end - start;

	if (length < 0)
	{
		strlcpy (out, "?empty filename?", 0);
	}
	else if (length == 0)
	{
		out[0] = 0;
	}
	else 
	{
		if (length > 31)
		{
			length = 31;
		}
		memcpy (out, start, length);
		out[length] = 0;
	}
}

/*
==================
COM_DefaultExtension
==================
*/
void COM_DefaultExtension (char *path, const char *extension, size_t len)
{
	const char	   *src;

	/*
	 * if path doesn't have a .EXT, append extension
	 * (extension should include the .)
	 */
	src = path + strlen (path) - 1;

	while (*src != '/' && src != path) 
	{
		// it has an extension
		if (*src == '.') return;
		src--;
	}
	strlcat (path, extension, len);
}

//============================================================================

/*
==============
COM_TokenLength

Safety measure
==============
*/
static int COM_TokenLength (int len)
{
	if (++len >= sizeof(com_token))
	{
		Host_Error ("COM_TokenLength: com_token overflow (max = %d)", sizeof(com_token));
	}
	return len;
}

/*
==============
COM_ParseToken

Parse a token out of a string
==============
*/
int COM_ParseToken(const char **datapointer, int returnnewline)
{
	int			len;
	const char	*data = *datapointer;

	len = 0;
	com_token[0] = '\0';

	if (!data)
	{
		*datapointer = '\0';
		return false;
	}

	// skip whitespace
skipwhite:
	// line endings:
	// UNIX: \n
	// Mac: \r
	// Windows: \r\n
	for (/**/; *data <= ' ' && ((*data != '\n' && *data != '\r') || !returnnewline); data++)
	{
		if (*data == 0)
		{
			// end of file
			*datapointer = '\0';
			return false;
		}
	}

	// handle Windows line ending
	if (data[0] == '\r' && data[1] == '\n')
	{
		data++;
	}

	if (data[0] == '/' && data[1] == '/')
	{
		// comment
		while (*data && *data != '\n' && *data != '\r')
		{
			data++;
		}
		goto skipwhite;
	}
	else if (data[0] == '/' && data[1] == '*')
	{
		// comment
		data++;

		while (*data && (data[0] != '*' || data[1] != '/'))
		{
			data++;
		}
		data += 2;
		goto skipwhite;
	}
	else if (*data == '\"')
	{
		// quoted string
		for (data++;*data != '\"';data++)
		{
			if (*data == '\\' && data[1] == '"' )
			{
				data++;
			}

			if (!*data || len >= (int)sizeof(com_token) - 1)
			{
				com_token[0] = '\0';
				*datapointer = '\0';
				return false;
			}
			com_token[len++] = *data;
		}
		com_token[len] = '\0';
		*datapointer = data+1;
		return true;
	}
	else if (*data == '\'')
	{
		// quoted string
		for (data++;*data != '\'';data++)
		{
			if (*data == '\\' && data[1] == '\'' )
			{
				data++;
			}

			if (!*data || len >= (int)sizeof(com_token) - 1)
			{
				com_token[0] = '\0';
				*datapointer = '\0';
				return false;
			}
			com_token[len++] = *data;
		}
		com_token[len] = '\0';
		*datapointer = data+1;
		return true;
	}
	else if (*data == '\r')
	{
		// translate Mac line ending to UNIX
		com_token[len++] = '\n';
		com_token[len] = '\0';
		*datapointer = data;
		return true;
	}
	else if (*data == '\n' || *data == '{' || *data == '}' || *data == ')' || *data == '(' || *data == ']' || *data == '[' || *data == '\'' || *data == ':' || *data == ',' || *data == ';')
	{
		// single character
		com_token[len++] = *data++;
		com_token[len] = '\0';
		*datapointer = data;
		return true;
	}
	else
	{
		// regular word
		for (/**/; *data > ' ' && *data != '{' && *data != '}' && *data != ')' && *data != '(' && *data != ']' && *data != '[' && *data != '\'' && *data != ':' && *data != ',' && *data != ';' && *data != '\'' && *data != '"'; data++)
		{
			if (len >= (int)sizeof(com_token) - 1)
			{
				com_token[0] = '\0';
				*datapointer = '\0';
				return false;
			}
			com_token[len++] = *data;
		}
		com_token[len] = '\0';
		*datapointer = data;
		return true;
	}
}

/*
==============
COM_ParseString

Parse a token out of a string
==============
*/
char *COM_ParseString (char *data)
{
	int	c;
	int	len;
	
	len = 0;
	com_token[0] = '\0';
	
	if (!data)
	{
		return '\0';
	}

// skip whitespace
skipwhite:
	while ( (c = *data) <= ' ')
	{
		if (c == 0)
		{
			return '\0';                    // end of file;
		}
		data++;
	}
	
// skip // comments
	if (c=='/' && data[1] == '/')
	{
		while (*data && *data != '\n')
		{
			data++;
		}
		goto skipwhite;
	}
	

// handle quoted strings specially
	if (c == '\"')
	{
		data++;

		while (1)
		{
			c = *data++;

			if (c=='\"' || !c)
			{
				com_token[len] = '\0';
				return data;
			}
			com_token[len] = c;
			len = COM_TokenLength (len);
		}
	}

// parse single characters
	if (c=='{' || c=='}'|| c==')'|| c=='(' || c=='\'' || c==':')
	{
		com_token[len] = c;
		len = COM_TokenLength (len);
		com_token[len] = '\0';
		return data+1;
	}

// parse a regular word
	do
	{
		com_token[len] = c;
		len = COM_TokenLength (len);
		data++;
		c = *data;
		if (c=='{' || c=='}'|| c==')'|| c=='(' || c=='\'' || c==':') break;
	} while (c>32);
	
	com_token[len] = 0;
	return data;
}

/*
================
COM_CheckParm

Returns the position (1 to argc-1) in the program's argument list
where the given parameter apears, or 0 if not present
================
*/
int COM_CheckParm (char *parm)
{
	int             i;
	
	for (i=1 ; i<com_argc ; i++)
	{
		if (!com_argv[i]) continue;               // NEXTSTEP sometimes clears appkit vars.

		if (!strcmp (parm, com_argv[i]))
		{
			return i;
		}
	}		
	return 0;
}

/*
================
COM_CheckRegistered

Looks for the pop.lmp file and verifies it.
Sets the "registered" cvar.
================
*/
void COM_CheckRegistered (void)
{
	filehandle_t	file;

	COM_FOpenFile ("gfx/pop.lmp", &file, FS_READ);

	if (file)
	{
		Cvar_Set ("cmdline", com_cmdline);
		Cvar_Set ("registered", "1");
		COM_FCloseFile (file);
	}
}

/*
================
COM_InitArgv
================
*/
void COM_InitArgv (int argc, char **argv)
{
	qboolean        safe;
	int             i, j, n;

// reconstitute the command line for the cmdline externally visible cvar
	n = 0;

	for (j=0 ; (j<MAX_NUM_ARGVS) && (j< argc) ; j++)
	{
		i = 0;

		while ((n < (CMDLINE_LENGTH - 1)) && argv[j][i])
		{
			com_cmdline[n++] = argv[j][i++];
		}

		if (n < (CMDLINE_LENGTH - 1))
		{
			com_cmdline[n++] = ' ';
		}
		else break;
	}
	com_cmdline[n] = 0;

	safe = false;

	for (com_argc=0 ; (com_argc<MAX_NUM_ARGVS) && (com_argc < argc) ; com_argc++)
	{
		largv[com_argc] = argv[com_argc];

		if (!strcmp ("-safe", argv[com_argc]))
		{
			safe = true;
		}
	}

	if (safe)
	{
		// force all the safe-mode switches. Note that we reserved extra space in
		// case we need to add these, so we don't need an overflow check
		for (i=0 ; i<NUM_SAFE_ARGVS ; i++)
		{
			largv[com_argc] = safeargvs[i];
			com_argc++;
		}
	}
	largv[com_argc] = argvdummy;
	com_argv = largv;

	if (COM_CheckParm ("-rogue"))
	{
		rogue = true;
		standard_quake = false;
	}

	if (COM_CheckParm ("-hipnotic"))
	{
		hipnotic = true;
		standard_quake = false;
	}
}

/*
================
COM_Init
================
*/
void COM_Init (char *basedir)
{
	byte    swaptest[2] = {1,0};

// set the byte swapping variables in a portable manner 
	if (*(short *)swaptest == 1)
	{
		bigendian = false;
		BigShort = ShortSwap;
		LittleShort = ShortNoSwap;
		BigLong = LongSwap;
		LittleLong = LongNoSwap;
		BigFloat = FloatSwap;
		LittleFloat = FloatNoSwap;
	}
	else
	{
		bigendian = true;
		BigShort = ShortNoSwap;
		LittleShort = ShortSwap;
		BigLong = LongNoSwap;
		LittleLong = LongSwap;
		BigFloat = FloatNoSwap;
		LittleFloat = FloatSwap;
	}
	Cvar_RegisterVariable (&registered);
	Cvar_RegisterVariable (&cmdline);

	Cmd_AddCommand ("path", COM_Path_f);
	Cmd_AddCommand ("gamedir", COM_Gamedir_f);

	COM_InitFilesystem ();
	COM_CheckRegistered ();
}


/*
============
va

does a varargs printf into a temp buffer, so I don't need to have
varargs versions of all text functions.
============
*/
char *va(const char *format, ...)
{
	// now cycles through 8 buffers to avoid problems in most cases
	va_list		argptr;
	static char string[8][1024], *s;
	static int	stringindex = 0;
	
	s = string[stringindex];
	stringindex = (stringindex + 1) & 7;
	va_start (argptr, format);
	va_vsnprintf ("va", s, sizeof(string[0]), format, argptr);
	va_end (argptr);
	
	return s;
}

/*
=============================================================================

QUAKE FILESYSTEM

=============================================================================
*/
#include <io.h>

#define PAK_HANDLES			0x400000	// insanity literally!!!
#define PAK_HASHSIZE		256
#define MAX_FILES_IN_PACK	5000

//
// in memory
//
typedef struct {
	char			name[MAX_QPATH];
	packmode_t		mode;
	FILE			*file;				// Only one of file or
	unzFile			*zip;				// zip will be used
} packHandle_t;

typedef struct
{
	char    name[MAX_QPATH];
	int     filepos, filelen;
	byte	hash;
} packfile_t;

typedef struct pack_s
{
	char		filename[MAX_OSPATH];
	short		hashoffset[PAK_HASHSIZE + 1];
	FILE		*pak;
	unzFile		*pk3;
	int         numfiles;
	packfile_t  *files;
} pack_t;

//
// on disk
//
typedef struct
{
	char    name[56];
	int     filepos, filelen;
} dpackfile_t;

typedef struct
{
	char    id[4];
	int     dirofs;
	int     dirlen;
} dpackheader_t;

typedef struct searchpath_s
{
	char				filename[MAX_OSPATH];
	pack_t				*pack;          // only one of filename / pack will be used
	struct searchpath_s *next;
} searchpath_t;

searchpath_t    *com_searchpaths;
searchpath_t	*com_base_searchpaths;
searchpath_t	*com_basesearchpaths;
packHandle_t	com_handles[PAK_HANDLES];
char			com_netpath[MAX_OSPATH];
char			com_basedir[MAX_OSPATH];
char			com_gamedir[MAX_OSPATH];
char			com_gamedirname[MAX_QPATH];
int				com_filesize;

/*
============
COM_Hash
============
*/
static byte COM_Hash (char *str)
{
	byte hash = '\0';
	
	while (*str)
	{
		hash += *str++;
	}
	return hash;
}

/*
=================
COM_GetFileByHandle

Returns a packHandle_t * for the given filehandle_t
=================
*/
packHandle_t *COM_GetFileByHandle (filehandle_t f)
{
	if (f <= 0 || f >= PAK_HANDLES)
	{
		Sys_Error ("COM_GetFileByHandle: %d is larger than %d", f, PAK_HANDLES);
	}
	return &com_handles[f-1];
}

/*
=================
COM_HandleForFile

Finds a free packHandle_t
=================
*/
packHandle_t *COM_HandleForFile (char *path, filehandle_t *f)
{
	packHandle_t	*handle;
	int				i;

	handle = com_handles;

	for (i = 0; i < PAK_HANDLES; i++, handle++)
	{
		if (!handle->file && !handle->zip)
		{
			memcpy(handle->name, path, sizeof(handle->name));

			*f = i + 1;

			return handle;
		}
	}

	// Failed
	return NULL;
}

/*
================
COM_FileLength
================
*/
int COM_FileLength (FILE *f)
{
	int		pos;
	int		end;

	pos = ftell (f);
	fseek (f, 0, SEEK_END);
	end = ftell (f);
	fseek (f, pos, SEEK_SET);

	return end;
}

/*
============
COM_Path_f
============
*/
void COM_Path_f (void)
{
	searchpath_t    *s;
	
	Con_Printf ("Current search path:\n");

	for (s = com_searchpaths ; s ; s = s->next)
	{
		if (s == com_base_searchpaths)
		{
			Con_Printf ("------------\n");
		}

		if (s->pack)
		{
			Con_Printf ("%s (%i files)\n", s->pack->filename, s->pack->numfiles);
		}
		else
		{
			Con_Printf ("%s\n", s->filename);
		}
	}
}

/*
============
COM_CreatePath

Only used for CopyFile
============
*/
void COM_CreatePath (char *path)
{
	char    *ofs;
	
	for (ofs = path + 1 ; *ofs ; ofs++)
	{
		if (*ofs == '/' || *ofs == '\\') 
		{      
			// create the directory
			*ofs = 0;
			Sys_mkdir (path);
			*ofs = '/';
		}
	}
}

/*
=================
COM_FOpenFileAppend

Returns file size or -1 on error
=================
*/
int COM_FOpenFileAppend (packHandle_t *handle)
{
	char	path[MAX_OSPATH];

	COM_CreatePath(handle->name);

	va_snprintf("COM_FOpenFileAppend", path, sizeof(path), "%s/%s", com_gamedir, handle->name);

	handle->file = fopen(path, "ab");

	if (handle->file)
	{
		if (developer.value)
		{
			Sys_Printf("file append: %s\n", path);
		}
		return COM_FileLength(handle->file);
	}

	if (developer.value)
	{
		Sys_Printf("file append: couldn't open %s\n", path);
	}
	return -1;
}

/*
============
COM_FOpenFileWrite

The filename will be prefixed by the current game directory
============
*/
int COM_FOpenFileWrite (packHandle_t *handle)
{
	char	path[MAX_OSPATH];

	COM_CreatePath(handle->name);

	va_snprintf("Com_FOpenFileWrite", path, sizeof(path), "%s/%s", com_gamedir, handle->name);

	handle->file = fopen(path, "wb");

	if (handle->file)
	{
		if (developer.value)
		{
			Sys_Printf("file write: %s\n", path);
		}
		return 0;
	}

	if (developer.value)
	{
		Sys_Printf("file write: couldn't open %s\n", path);
	}
	return -1;
}

/*
===========
COM_FOpenFileRead

Finds the file in the search path.
Sets com_filesize and one of handle or file
===========
*/
int COM_FOpenFileRead (packHandle_t *handle)
{
	searchpath_t    *search;
	pack_t          *pak;
	int             i;
	byte			hash;

	// build a hash for it.
	hash = COM_Hash (handle->name);
//
// search through the path, one element at a time
//
	for (search = com_searchpaths ; search ; search = search->next)
	{
		// is the element a pak file?
		if (search->pack)
		{
			int  start, stop;

			// look through all the pak file elements
			pak = search->pack;
			start = pak->hashoffset[hash];
            stop = pak->hashoffset[hash + 1];

			// pack is a pack
			for (i = start ; i < stop ; i++)
			{ 
				if (!strcasecmp (pak->files[i].name,  handle->name))
				{       
					// found it!
					Sys_Printf ("file open: %s : %s\n", pak->filename,  handle->name);

					// open a new file on the pakfile
					if (pak->pak)
					{	
						// PAK
						handle->file = fopen(pak->filename, "rb");

						if (handle->file)
						{
							if (fseek(handle->file, pak->files[i].filepos, SEEK_SET) != -1)
							{
								com_filesize = pak->files[i].filelen;

								va_snprintf ("COM_FOpenFileRead", com_netpath, sizeof(com_netpath), "%s#%i", pak->filename, i);

								return com_filesize;
							}
						}
					}
					else if (pak->pk3)
					{
						handle->zip = unzOpen(pak->filename);

						if (handle->zip)
						{
							if (unzLocateFile(handle->zip, handle->name, 2) == UNZ_OK)
							{
								if (unzOpenCurrentFile(handle->zip) == UNZ_OK)
								{
									com_filesize = pak->files[i].filelen;

									va_snprintf ("COM_FOpenFileRead", com_netpath, sizeof(com_netpath), "%s#%i", pak->filename, i);

									return com_filesize;
								}
							}
							unzClose(handle->zip);
						}
					}
					Sys_Error("Couldn't reopen %s", pak->filename);
				}
			}
		}
		else
		{
			//expand filename to absolute path
			va_snprintf ("COM_FOpenFileRead", com_netpath, sizeof(com_netpath), "%s/%s", search->filename, handle->name);

			if (developer.value)
			{
				Sys_Printf ("file open: %s\n", com_netpath);
			}

			if (!(handle->file = fopen(com_netpath, "rb"))) continue;

			com_filesize = COM_FileLength(handle->file);

			return com_filesize;
		}		
	}	

	// found nothing
	handle->file = NULL;
	handle->zip = NULL;
	com_filesize = -1;
	com_netpath[0] = '\0';
	
	if (developer.value)
	{
		Sys_Printf("file open: couldn't find %s\n", handle->name);
	}
	return -1;
}

/*
=================
COM_FOpenDemoFile

Old legacy junk will have to do for now though.
=================
*/
int COM_FOpenDemoFile (char *filename, FILE **file)
{
	searchpath_t	*search;
	pack_t			*pak;
	int				i;

	// search through the path, one element at a time
	for (search = com_searchpaths ; search ; search = search->next)
	{
		// is the element a pak file?
		if (search->pack)
		{
			// look through all the pak file elements
			pak = search->pack;

			for (i=0 ; i<pak->numfiles ; i++)
			{
				if (!strcmp(pak->files[i].name, filename))	// found it!
				{
					if (developer.value)
					{
						Sys_Printf ("file open demo: %s : %s\n", pak->filename, filename);
					}

					// open a new file on the pakfile
					if (!(*file = fopen(pak->filename, "rb")))
					{
						Sys_Error ("file open demo: Couldn't reopen %s", pak->filename);
					}
					fseek (*file, pak->files[i].filepos, SEEK_SET);
					com_filesize = pak->files[i].filelen;

					va_snprintf ("COM_FOpenDemoFile", com_netpath, sizeof(com_netpath), "%s#%i", pak->filename, i);
					return com_filesize;
				}
			}
		}
		else
		{               
			// check a file in the directory tree
			va_snprintf ("COM_FOpenDemoFile", com_netpath, sizeof(com_netpath), "%s/%s", search->filename, filename);

			if (!(*file = fopen(com_netpath, "rb"))) continue;

			if (developer.value)
			{
				Sys_Printf ("file open demo: %s\n", com_netpath);
			}
			com_filesize = COM_FileLength (*file);

			return com_filesize;
		}
	}

	// found nothing
	*file = NULL;
	com_filesize = -1;
	com_netpath[0] = '\0';

	if (developer.value)
	{
		Sys_Printf ("file open demo: can't find %s\n", filename);
	}
	return -1;
}

/*
=================
COM_FOpenFile

Opens a file for "mode".
Returns file size or -1 if an error occurs/not found.
Can open separate files as well as files inside pack files (both PAK 
and PK3).
=================
*/
int COM_FOpenFile (char *name, filehandle_t *f, packmode_t mode)
{
	packHandle_t	*handle;
	int				size;

	// get handle
	handle = COM_HandleForFile(name, f);

	if (!handle)
	{
		Sys_Error("COM_FOpenFile: no free handles");
	}
	memcpy(handle->name, name, sizeof(handle->name));

	handle->mode = mode;

	switch (mode)
	{
	case FS_READ:
		size = COM_FOpenFileRead(handle);
		break;
	case FS_WRITE:
		size = COM_FOpenFileWrite(handle);
		break;
	case FS_APPEND:
		size = COM_FOpenFileAppend(handle);
		break;
	default:
		Sys_Error("COM_FOpenFile: bad mode (%i)", mode);
	}

	if (size != -1)
	{
		return size;
	}

	// Couldn't open, so free the handle
	memset(handle, 0, sizeof(*handle));

	*f = 0;

	return -1;
}

/*
=================
COM_FCloseFile
=================
*/
void COM_FCloseFile (filehandle_t f)
{
	packHandle_t *handle;

	// get handle
	handle = COM_GetFileByHandle(f);

	if (handle->file)
	{
		fclose(handle->file);
	}
	else if (handle->zip) 
	{
		unzCloseCurrentFile(handle->zip);
		unzClose(handle->zip);
	}
	memset(handle, 0, sizeof(*handle));
}

/*
=================
COM_FSeek
=================
*/
void COM_FSeek (filehandle_t f, int offset, packorigin_t origin)
{
	packHandle_t	*handle;
	unz_file_info	info;
	int				remaining, r, len;
	byte			dummy[0x8000];

	// get handle
	handle = COM_GetFileByHandle(f);

	if (handle->file)
	{
		switch (origin)
		{
		case FS_SEEK_SET:
			fseek(handle->file, offset, SEEK_SET);
			break;
		case FS_SEEK_CUR:
			fseek(handle->file, offset, SEEK_CUR);
			break;
		case FS_SEEK_END:
			fseek(handle->file, offset, SEEK_END);
			break;
		default:
			Sys_Error("COM_FSeek: bad origin (%i)", origin);
		}
	}
	else if (handle->zip)
	{
		switch (origin)
		{
		case FS_SEEK_SET:
			remaining = offset;
			break;
		case FS_SEEK_CUR:
			remaining = offset + unztell(handle->zip);
			break;
		case FS_SEEK_END:
			unzGetCurrentFileInfo(handle->zip, &info, NULL, 0, NULL, 0, NULL, 0);
			remaining = offset + info.uncompressed_size;
			break;
		default:
			Sys_Error("COM_FSeek: bad origin (%i)", origin);
		}

		// Reopen the file
		unzCloseCurrentFile(handle->zip);
		unzOpenCurrentFile(handle->zip);

		// Skip until the desired offset is reached
		while (remaining)
		{
			len = remaining;

			if (len > sizeof(dummy))
			{
				len = sizeof(dummy);
			}
			r = unzReadCurrentFile(handle->zip, dummy, len);

			if (r <= 0) break;

			remaining -= r;
		}
	}
}

/*
=================
COM_FTell

Returns -1 if an error occurs
=================
*/
int COM_FTell (filehandle_t f)
{
	int				pos;
	packHandle_t	*handle;

	// get handle
	handle = COM_GetFileByHandle(f);

	if (handle->file)
	{
		pos = ftell(handle->file);
	}
	else if (handle->zip)
	{
		pos = unztell(handle->zip);
	}
	return pos;
}

/*
=================
COM_FGetc -> fgetc

utility functions to avoid 
having to rewrite the entire engine.
=================
*/
int COM_FGetc (filehandle_t f)
{
	byte	x[1];
	int		ret;

	ret = COM_FRead(x, sizeof(byte), 1, f);

	if (ret == 0)
	{
		return (EOF);
	}
	return ((int)x[0]);
}

/*
=================
COM_FGetcLittleShort

utility functions to avoid 
having to rewrite the entire engine.
=================
*/
int COM_FGetcLittleShort (filehandle_t f)
{
	byte	b1, b2;

	b1 = COM_FGetc(f);
	b2 = COM_FGetc(f);

	return (short)(b1 + b2 * 256);
}

/*
=================
COM_FGetcLittleLong

utility functions to avoid 
having to rewrite the entire engine.
=================
*/
int COM_FGetcLittleLong (filehandle_t f)
{
	byte	b1, b2, b3, b4;

	b1 = COM_FGetc(f);
	b2 = COM_FGetc(f);
	b3 = COM_FGetc(f);
	b4 = COM_FGetc(f);

	return b1 + (b2<<8) + (b3<<16) + (b4<<24);
}

/*
=================
COM_FRead

Properly handles partial reads of size up to count times.
No error if it can't read.
=================
*/
int COM_FRead (void *buffer, int size, int count, filehandle_t f)
{
	packHandle_t	*handle;
	int				loops, remaining, r;
	byte			*buf;
	qboolean		tried = false;

	// get handle
	handle = COM_GetFileByHandle(f);

	// Read
	loops = count;

	//remaining = size;
	buf = (byte *)buffer;

	while (loops)
	{	
		// Read in chunks
		remaining = size;

		while (remaining)
		{
			if (handle->file)
			{
				r = fread(buf, 1, remaining, handle->file);
			}
			else if (handle->zip)
			{
				r = unzReadCurrentFile(handle->zip, buf, remaining);
			}
			else
			{
				Sys_Printf("COM_FRead: couldnt read from %s\n", handle->name);
				return 0;
			}

			if (r == 0)
			{
				if (!tried)
				{	
					// We might have been trying to read from a CD
					Sys_Printf("COM_FRead: tried to read from %s\n", handle->name);
					CDAudio_Stop();
					tried = true;
				}
				else 
				{
					Sys_Printf("COM_FRead: 0 bytes read from %s\n", handle->name);
					return size - remaining;
				}
			}
			else if (r == -1)
			{
				Sys_Error("COM_FRead: -1 bytes read from %s", handle->name);
			}
			remaining -= r;
			buf += r;
		}
		loops--;
	}
	return size;
}

/*
=================
COM_Read

Properly handles partial reads.
=================
*/
int COM_Read (void *buffer, int size, filehandle_t f)
{
	packHandle_t	*handle;
	int				remaining, r;
	byte			*buf;
	qboolean		tried = false;

	// get handle
	handle = COM_GetFileByHandle(f);

	// Read
	remaining = size;
	buf = (byte *)buffer;

	while (remaining)
	{
		if (handle->file)
		{
			r = fread(buf, 1, remaining, handle->file);
		}
		else if (handle->zip)
		{
			r = unzReadCurrentFile(handle->zip, buf, remaining);
		}
		else
		{
			Sys_Printf("COM_FRead: couldnt read from %s\n", handle->name);
			return 0;
		}

		if (r == 0)
		{
			if (!tried)
			{	
				// We might have been trying to read from a CD
				Sys_Printf("COM_FRead: tried to read from %s\n", handle->name);
				CDAudio_Stop();
				tried = true;
			}
			else
			{	
				// Already tried once
				Sys_Printf("COM_Read: 0 bytes read from %s\n", handle->name);
				return size - remaining;
			}
		}
		else if (r == -1)
		{
			Sys_Error("COM_Read: -1 bytes read from %s", handle->name);
		}
		remaining -= r;
		buf += r;
	}
	return size;
}

/*
============
COM_LoadFile

Filename are reletive to the quake directory.
Allways appends a 0 byte.
============
*/
static byte *COM_LoadFile (char *path, int usehunk)
{
	filehandle_t	handle;
	byte			*buf;
	char			base[32];
	int				size;

    // quiet compiler warning
	buf = NULL;

	// look for it in the filesystem or pack files
	size = COM_FOpenFile (path, &handle, FS_READ);

	if (size == -1 || size == 0)
	{
		return NULL;
	}

	// extract the filename base name for hunk tag
	COM_FileBase (path, base);

	switch (usehunk)
	{
	case 0:
		// main memory alloc
		buf = Heap_MemAlloc (size + 2);
		break;
	case 1:
		// permanent hunk (overwritten at next map start)
		buf = Heap_ZAlloc (size + 2);
		break;
	case 2:
		// temporary hunk (overwritten at next tempalloc)
		buf = Heap_ScratchZAlloc (size + 2);
		break;
	default:
		Sys_Error ("COM_LoadFile: bad usehunk");
		break;
	}

	if (!buf)
	{
		Sys_Error ("COM_LoadFile: not enough space for %s (%dk)", path, (size + 1) / 1024);
	}
	((byte *)buf)[size] = 0;

	if (COM_Read(buf, size, handle) != size)
	{
		Sys_Error ("COM_LoadFile: error reading %s", path);
	}
	COM_FCloseFile (handle);

	return buf;
}

byte *COM_LoadHeapFile (char *path)
{
	return COM_LoadFile (path, 0);
}

byte *COM_LoadHunkFile (char *path)
{
	return COM_LoadFile (path, 1);
}

byte *COM_LoadTempFile (char *path)
{
	return COM_LoadFile (path, 2);
}

static int COM_PakComp (const void *arg1, const void *arg2)
{
	return ((packfile_t *)arg1)->hash - ((packfile_t *)arg2)->hash;
}

/*
=================
COM_LoadPackFile

Takes an explicit (not game tree related) path to a pak file.

Loads the header and directory, adding the files at the beginning
of the list so they override previous pack files.
=================
*/
pack_t *COM_LoadPackFile (char *packfile)
{
	dpackheader_t   header;
	int             i;
	int				dofs, dlen;
	packfile_t      *newfiles;
	int             numpackfiles;
	pack_t          *pack;
	FILE            *packhandle;
	dpackfile_t     info[MAX_FILES_IN_PACK];

	packhandle = fopen(packfile, "rb");

	if (!packhandle)
	{
		return NULL;
	}
	fread (&header, 1, sizeof(header), packhandle);

	if (memcmp(header.id, "PACK", 4))
	{
		Sys_Error ("%s is not a packfile", packfile);
	}
	dofs = header.dirofs = LittleLong (header.dirofs);
	dlen = header.dirlen = LittleLong (header.dirlen);

	if (dofs < sizeof(header) || dlen < sizeof(dpackfile_t))
	{
		Sys_Error ("Directory is outside range (%d-%d, pak = %d-%d) in packfile %s", dofs, dofs + dlen - 1, sizeof(header), packfile);
	}
	numpackfiles = header.dirlen / sizeof(dpackfile_t);

	if (numpackfiles > MAX_FILES_IN_PACK)
	{
		Sys_Error ("Packfile %s has too many files (%i, max = %i)", packfile, numpackfiles, MAX_FILES_IN_PACK);
	}
	newfiles = Heap_MemAlloc(numpackfiles * sizeof(packfile_t));

	fseek (packhandle, header.dirofs, SEEK_SET);
	fread (&info, 1, header.dirlen, packhandle);
	
	// parse the directory
	for (i=0 ; i<numpackfiles ; i++)
	{
		int ofs = newfiles[i].filepos = LittleLong(info[i].filepos);
		int len = newfiles[i].filelen = LittleLong(info[i].filelen);
		
		info[i].name[sizeof(info[i].name) - 1] = '\0'; // Make sure it's terminated
		strlcpy_s (newfiles[i].name, info[i].name);
		
		// In pak0 of chapters_gom, a dir entry is (0,0), so allow 0 sized bad entries
		if (ofs < 0 || len < 0 || len > 0 &&																	// Entry has a size
		   (ofs < sizeof(header) ||																				// Entry start is inside header
		    ofs >= dofs && ofs < dofs + dlen ||																	// Entry start is inside directory
		    ofs + len > dofs && ofs + len <= dofs + dlen ||														// Entry end is inside directory
		    ofs < dofs && ofs + len > dofs + dlen))																// Directory is inside entry
			Sys_Error ("File %s is outside range (%d-%d, pak = %d-%d, dir = %d-%d) in packfile %s", newfiles[i].name, ofs, ofs + len - 1, dofs, dofs + dlen - 1, packfile);
	}
	pack = Heap_MemAlloc (sizeof (pack_t));
	strlcpy_s (pack->filename, packfile);
	pack->pak = packhandle;
	pack->pk3 = NULL;
	pack->numfiles = numpackfiles;
	pack->files = newfiles;
	
	// Generate hash offsets to speed file searches
	for (i = 0; i < numpackfiles; ++i)
	{
		newfiles[i].hash = COM_Hash (newfiles[i].name);
	}

	// Sort in ascending order
	qsort (newfiles, numpackfiles, sizeof(packfile_t), COM_PakComp);

	for (i = 0; i < PAK_HASHSIZE; ++i)
	{
		pack->hashoffset[i] = -1;
	}
	pack->hashoffset[i] = numpackfiles;

	// Find first file index with a certain hash
	for (i = numpackfiles - 1; i >= 0; --i)
	{
		pack->hashoffset[newfiles[i].hash] = i;
	}

	// Propagate real offsets downwards in array
	for (i = PAK_HASHSIZE - 1; i >= 0; --i)
	{
		if (pack->hashoffset[i] == -1)
		{
			pack->hashoffset[i] = pack->hashoffset[i + 1];
		}
	}
	Con_Printf ("Added packfile %s (%i files)\n", packfile, numpackfiles);
	return pack;
}

/*
=================
COM_LoadPK3File

Takes an explicit (not game tree related) path to a pack file.

Loads the header and directory, adding the files at the beginning of
the list so they override previous pack files.
=================
*/
pack_t *COM_LoadPK3File (char *packfile)
{
	int				numpackfiles, i = 0;
	packfile_t		*newfiles;
	pack_t			*pack;
	unzFile			*packhandle;
	unz_global_info	global;
	unz_file_info	info;
	int				status;
	char			fileName[MAX_QPATH];

	packhandle = unzOpen(packfile);

	if (!packhandle)
	{
		return NULL;
	}

	if (unzGetGlobalInfo(packhandle, &global) != UNZ_OK)
	{
		unzClose(packhandle);
		Sys_Error("COM_LoadPK3File: %s is not a pack file", packfile);
	}
	numpackfiles = global.number_entry;

	if (numpackfiles > MAX_FILES_IN_PACK || numpackfiles == 0)
	{
		unzClose(packhandle);
		Sys_Error("COM_LoadPK3File: %s has %i files", packfile, numpackfiles);
	}
	newfiles = Heap_MemAlloc(numpackfiles * sizeof(packfile_t));

	// Parse the directory
	status = unzGoToFirstFile(packhandle);

	while (status == UNZ_OK)
	{
		fileName[0] = '\0';
		unzGetCurrentFileInfo(packhandle, &info, fileName, MAX_QPATH, NULL, 0, NULL, 0);
		strlcpy_s(newfiles[i].name, fileName);
		newfiles[i].filepos = -1;		// Not used in ZIP files
		newfiles[i].filelen = info.uncompressed_size;
		i++;

		status = unzGoToNextFile(packhandle);
	}
	pack = Heap_MemAlloc (sizeof (pack_t));
	strlcpy_s(pack->filename, packfile);
	pack->pak = NULL;
	pack->pk3 = packhandle;
	pack->numfiles = numpackfiles;
	pack->files = newfiles;

	// Generate hash offsets to speed file searches
	for (i = 0; i < numpackfiles; ++i)
	{
		newfiles[i].hash = COM_Hash (newfiles[i].name);
	}

	// Sort in ascending order
	qsort (newfiles, numpackfiles, sizeof(packfile_t), COM_PakComp);

	for (i = 0; i < PAK_HASHSIZE; ++i)
	{
		pack->hashoffset[i] = -1;
	}
	pack->hashoffset[i] = numpackfiles;

	// Find first file index with a certain hash
	for (i = numpackfiles - 1; i >= 0; --i)
	{
		pack->hashoffset[newfiles[i].hash] = i;
	}

	// Propagate real offsets downwards in array
	for (i = PAK_HASHSIZE - 1; i >= 0; --i)
	{
		if (pack->hashoffset[i] == -1)
		{
			pack->hashoffset[i] = pack->hashoffset[i + 1];
		}
	}
	Con_Printf ("Added packfile %s (%i files)\n", packfile, numpackfiles);
	return pack;
}

/*
================
COM_AddGameDirectory

Sets com_gamedir, adds the directory to the head of the path,
then loads and adds pak1.pak pak2.pak ... 
================
*/
void COM_AddGameDirectory (char *dir)
{
	int             i;
	searchpath_t    *search;
	pack_t          *pak;
	char            pakfile[MAX_OSPATH], *p;

	if ((p = strrchr(dir, '/')))
	{
		strlcpy (com_gamedirname, ++p, sizeof(com_gamedirname));
	}
	else
	{
		strlcpy (com_gamedirname, p, sizeof(com_gamedirname));
	}
	strlcpy (com_gamedir, dir, sizeof(com_gamedir));

	// add the directory to the search path
	search = Heap_MemAlloc (sizeof (searchpath_t));
	strlcpy (search->filename, dir, sizeof(search->filename));
	search->pack = NULL;
	search->next = com_searchpaths;
	com_searchpaths = search;

	// add any pak files in the format pak0.pak pak1.pak, ...
	for (i = 0; ; i++)
	{
		va_snprintf ("AddGameDirectory", pakfile, sizeof(pakfile), "%s/pak%i.pak", dir, i);

		pak = COM_LoadPackFile (pakfile);

		if (!pak) break;

		search = Heap_MemAlloc (sizeof (searchpath_t));
		strlcpy (search->filename, dir, sizeof(search->filename));
		search->pack = pak;
		search->next = com_searchpaths;
		com_searchpaths = search;               
	}

	// add any pk3 files in the format pak0.pk3 pak1.pk3, ...
	for (i = 0; ; i++)
	{
		va_snprintf ("AddGameDirectory", pakfile, sizeof(pakfile), "%s/pak%i.pk3", dir, i);

		pak = COM_LoadPK3File (pakfile);

		if (!pak) break;

		search = Heap_MemAlloc (sizeof (searchpath_t));
		strlcpy (search->filename, dir, sizeof(search->filename));
		search->pack = pak;
		search->next = com_searchpaths;
		com_searchpaths = search;               
	}
}

/*
================
COM_SetGameDir

Sets the gamedir and path to a different directory.
================
*/
void COM_SetGameDir (char *dir)
{
	int				i;
	searchpath_t	*search, *next;
	pack_t			*pack;
	char			pakfile[MAX_OSPATH];

	if (strstr(dir, "..") || strstr(dir, "/") || strstr(dir, "\\") || strstr(dir, ":"))
	{
		Sys_Printf ("Gamedir should be a single filename, not a path\n");
		return;
	}

	if (!strcasecmp(com_gamedirname, dir)) return;		// still the same

	strlcpy (com_gamedirname, dir, sizeof(com_gamedirname));

	// free up any current game dir info
	while (com_searchpaths != com_base_searchpaths)
	{
		if (com_searchpaths->pack)
		{
			pack = com_searchpaths->pack;

			if (pack->pak)
			{
				fclose (com_searchpaths->pack->pak);
			}

			if (pack->pk3)
			{
				unzClose(com_searchpaths->pack->pk3);
			}
			Heap_MemAllocFree(pack->files);
			Heap_MemAllocFree(pack);
		}
		next = com_searchpaths->next;
		Heap_MemAllocFree(com_searchpaths);
		com_searchpaths = next;
	}
	va_snprintf ("COM_SetGameDir", com_gamedir, sizeof(com_gamedir), "%s/%s", com_basedir, dir);

	// skip id1
	if (!strcasecmp(dir, "id1")) return;

	// add the directory to the search path
	search = Heap_MemAlloc (sizeof(searchpath_t));
	strlcpy (search->filename, com_gamedir, sizeof(search->filename));
	search->pack = NULL;
	search->next = com_searchpaths;
	com_searchpaths = search;

	// add any pak files in the format pak0.pak pak1.pak, ...
	for (i=0 ; ; i++)
	{
		va_snprintf ("COM_SetGameDir", pakfile, sizeof(pakfile), "%s/pak%i.pak", com_gamedir, i);
		if (!(pack = COM_LoadPackFile(pakfile)))	break;
		search = Heap_MemAlloc (sizeof(searchpath_t));
		search->pack = pack;
		search->next = com_searchpaths;
		com_searchpaths = search;               
	}

	// add any pk3 files in the format pak0.pk3 pak1.pk3, ...
	for (i = 0; ; i++)
	{
		va_snprintf ("COM_SetGameDir", pakfile, sizeof(pakfile), "%s/pak%i.pk3", com_gamedir, i);
		if (!(pack = COM_LoadPK3File(pakfile))) break;
		search = Heap_MemAlloc (sizeof(searchpath_t));
		search->pack = pack;
		search->next = com_searchpaths;
		com_searchpaths = search; 
	}
}

/*
================
COM_Gamedir_f

Sets the gamedir and path to a different directory.
================
*/
void COM_Gamedir_f (void)
{
	char	*dir;

	if (cmd_source != src_command) return;

	if (Cmd_Argc() == 1)
	{
		Sys_Printf ("Current gamedir: %s\n", com_gamedirname);
		return;
	}

	if (Cmd_Argc() != 2)
	{
		Sys_Printf ("Usage: gamedir <newdir>\n");
		return;
	}
	dir = Cmd_Argv (1);

	if (strstr(dir, "..") || strstr(dir, "/") || strstr(dir, "\\") || strstr(dir, ":") )
	{
		Sys_Printf ("gamedir should be a single filename, not a path\n");
		return;
	}

	if (cls.state != ca_disconnected)
	{
		Sys_Printf ("you must disconnect before changing gamedir\n");
		return;
	}
	COM_SetGameDir (dir);
}

/*
====================================
COM_SearchWildCards

not used atm will be later as its superior to the old findstring.
takes a searchstring with wildcards (*) and searches for all files maching it.
returns an array of strings that need to be cleared.
====================================
*/
int COM_SearchWildCards (char *searchstring, char **found, int maxlen, qboolean fullpath)
{
	int				i;
	searchpath_t	*search;
	char			basestring[MAX_QPATH];
	int				baselen, numfound = 0;

	baselen = (int)COM_SkipPath(searchstring) - (int)searchstring;

	strlcpy(basestring, searchstring, baselen);

	for (search = com_searchpaths ; search; search = search->next)
	{
		if (search->pack)
		{
			// look through all the pak file elements
			pack_t *pak = search->pack;

			if(fullpath) continue;

			for (i=0 ; i<pak->numfiles && numfound < maxlen ; i++)
			{
				if (!strcmp (basestring, pak->files[i].name))
				{       
					// found a matching mapname.. now to match wildcards.
					char	*substr = pak->files[i].name;
					char	*wildcardstring = COM_SkipPath(searchstring);
					int		n = 0; //postition in wildcardstring

					while(1)
					{
						char	tmp[MAX_QPATH];
						int		s = 0;

						switch(wildcardstring[n])
						{
						case '\0':
							if(substr[0])
							{
								goto next;
							}
							else
							{
								//Found it!
								found[numfound] = Heap_MemAlloc(strlen(pak->files[i].name) * sizeof(char));
								strcpy(found[numfound], pak->files[i].name);
								numfound++;
							}

						case '*':
							while(wildcardstring[n] != '\0' && wildcardstring[n] != '*' && wildcardstring[n] != '?')
							{
								tmp[s++] = substr[n++];
							}
							tmp[s] = '\0';
							substr = strstr(tmp, substr);

							if(!substr)
							{
								goto next;
							}
						default:
							while(wildcardstring[n] != '\0' && wildcardstring[n] != '*' && wildcardstring[n] != '?')
							{
								tmp[s++] = substr[n++];
							}
							tmp[s] = '\0';

							if(strcmp(tmp, substr))
							{
								goto next;
							}
							substr += s-1;
						}
						n++;
						substr++;
					}
					found[numfound] = Heap_MemAlloc(strlen(pak->files[i].name) * sizeof(char));
					strcpy(found[numfound], pak->files[i].name);
					numfound++;
next:;
				}
			}
		}
		else
		{               
			char				filestring[MAX_QPATH];
			char				dirstring [MAX_QPATH];
			struct _finddata_t	fileinfo;
			int					handle;

			va_snprintf ("COM_SearchWildCards", dirstring, sizeof(dirstring), "%s/%s", search->filename, searchstring);
			handle = _findfirst (dirstring, &fileinfo);

			if (handle != -1)
			{
				do
				{
					if (fileinfo.name[0] == '.') continue;

					if(fullpath)
					{
						va_snprintf("COM_SearchWildCards", filestring, sizeof(filestring), "%s/%s%s", search->filename, basestring, fileinfo.name);
					}
					else
					{
						va_snprintf("COM_SearchWildCards", filestring, sizeof(filestring), "%s%s", basestring, fileinfo.name);
					}
					found[numfound] = Heap_MemAlloc(strlen(filestring) * sizeof(char));
					strcpy(found[numfound], filestring);
					numfound++;
				} while (_findnext( handle, &fileinfo ) != -1 && numfound < maxlen);
				_findclose (handle);
			}
		}
	}
	return numfound;
}

/*
================
COM_InitFilesystem
================
*/
void COM_InitFilesystem (void)
{
	int				i;
	pack_t			*pack;
	searchpath_t	*search, *next;

	Sys_Printf("\n----- Filesystem Initialization -----\n");

	// -basedir <path>
	// Overrides the system supplied base directory (under GAMENAME)
	if ((i = COM_CheckParm("-basedir")) && i + 1 < com_argc)
	{
		strlcpy (com_basedir, com_argv[i + 1], sizeof(com_basedir));
	}
	else
	{
		strlcpy (com_basedir, host_parms.basedir, sizeof(com_basedir));
	}

	for (i=0 ; i < (int)strlen(com_basedir) ; i++)
	{
		if (com_basedir[i] == '\\')
		{
			com_basedir[i] = '/';
		}
	}
	i = strlen(com_basedir) - 1;

	if (i >= 0 && com_basedir[i] == '/')
	{
		com_basedir[i] = '\0';
	}

	// free up any current game dir info
	while (com_searchpaths != com_base_searchpaths)
	{
		if (com_searchpaths->pack)
		{
			pack = com_searchpaths->pack;

			if (pack->pak)
			{
				fclose (pack->pak);
			}

			if (pack->pk3)
			{
				unzClose(pack->pk3);
			}
			Heap_MemAllocFree(pack->files);
			Heap_MemAllocFree(pack);
		}
		next = com_searchpaths->next;
		Heap_MemAllocFree(com_searchpaths);
		com_searchpaths = next;
	}

	// start up with GAMENAME by default (id1)
	COM_AddGameDirectory (va("%s/"GAMENAME, com_basedir));

	if (COM_CheckParm ("-rogue"))
	{
		COM_AddGameDirectory (va("%s/rogue", com_basedir));
	}

	if (COM_CheckParm ("-hipnotic"))
	{
		COM_AddGameDirectory (va("%s/hipnotic", com_basedir));
	}

	// any set gamedirs will be freed up to here
	com_base_searchpaths = com_searchpaths;

	// -game <gamedir>
	// Adds basedir/gamedir as an override game
	if ((i = COM_CheckParm("-game")) && i + 1 < com_argc)
	{
		COM_AddGameDirectory (va("%s/%s", com_basedir, com_argv[i+1]));
	}

	// -path <dir or packfile> [<dir or packfile>] ...
	// Fully specifies the exact search path, overriding the generated one
	if ((i = COM_CheckParm ("-path")))
	{
		com_searchpaths = '\0';

		while (++i < com_argc)
		{
			if (!com_argv[i] || com_argv[i][0] == '+' || com_argv[i][0] == '-')	break;
			
			search = Heap_MemAlloc (sizeof (searchpath_t));

			if (!strcasecmp (COM_FileExtension (com_argv[i]), "pak"))
			{
				search->pack = COM_LoadPackFile (com_argv[i]);

				if (!search->pack) 
				{
					Sys_Printf ("Couldn't load packfile: %s", com_argv[i]);
					return;
				}
			}
			else if (!strcasecmp (COM_FileExtension (com_argv[i]), "pk3"))
			{
				search->pack = COM_LoadPK3File (com_argv[i]);

				if (!search->pack) 
				{
					Sys_Printf ("Couldn't load pk3file: %s", com_argv[i]);
					return;
				}
			}
			else
			{
				strlcpy_s (search->filename, com_argv[i]);
			}
			search->next = com_searchpaths;
			com_searchpaths = search;
		}
	}
}

/*
=================
COM_ShutdownFileSystem
=================
*/
void COM_ShutdownFileSystem (void)
{
	packHandle_t	*handle;
	searchpath_t	*next;
	pack_t			*pack;
	int				i;

	// Close all files
	for (i = 0, handle = com_handles; i < PAK_HANDLES; i++, handle++)
	{
		if (handle->file)
		{
			fclose(handle->file);
		}

		if (handle->zip)
		{
			unzCloseCurrentFile(handle->zip);
			unzClose(handle->zip);
		}
	}

	// Free the search paths
	while (com_searchpaths)
	{
		if (com_searchpaths->pack)
		{
			pack = com_searchpaths->pack;

			if (pack->pak)
			{
				fclose(pack->pak);
			}

			if (pack->pk3)
			{
				unzClose(pack->pk3);
			}
			Heap_MemAllocFree(pack->files);
			Heap_MemAllocFree(pack);
		}
		next = com_searchpaths->next;
		Heap_MemAllocFree(com_searchpaths);
		com_searchpaths = next;
	}
}

