/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __MODEL__
#define __MODEL__

#include "modelgen.h"
#include "spritegn.h"

/*

d*_t structures are on-disk representations
m*_t structures are in-memory

*/

// entity effects

#define	EF_BRIGHTFIELD			1
#define	EF_MUZZLEFLASH 			2
#define	EF_BRIGHTLIGHT 			4
#define	EF_DIMLIGHT 			8


/*
==============================================================================

BRUSH MODELS

==============================================================================
*/


//
// in memory representation
//
// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct
{
	vec3_t		position;
} mvertex_t;

#define	SIDE_FRONT	0
#define	SIDE_BACK	1
#define	SIDE_ON		2


// plane_t structure
// !!! if this is changed, it must be changed in asm_i386.h too !!!
typedef struct mplane_s
{
	vec3_t	normal;
	float	dist;
	byte	type;			// for texture axis selection and fast side tests
	byte	signbits;		// signx + signy<<1 + signz<<1
	byte	pad[2];
} mplane_t;

#define LIQUIDTYPE_NOTLIQUID	0
#define LIQUIDTYPE_LAVA			1
#define LIQUIDTYPE_WATER		2
#define LIQUIDTYPE_SLIME		3
#define LIQUIDTYPE_TELE			4

// moved from gl_draw.c
// all structs containing a texture number will now point to this instead
typedef struct gltexture_s
{
	unsigned int			gltexnum;
	char					identifier[64];
	int						width, height;
	byte					*data;

	byte					md5digest[16];
	int						textureage;

	// additional texture numbers
	unsigned int			lumatexnum;
	unsigned int			normtexnum;

	int						liquidtype;

	// store 4 colours so that we can pass this through to glFogfv
	float					rgb[4];

	// linked list of all active textures
	// moved the chains from below to here so that it can be used for chaining and/or listing
	// anything with a texture on it.
	struct		gltexture_s *TextureList;
	struct		msurface_s	*texturechain;

	// flags list defined in gl_state.h
	unsigned int			flags;
} gltexture_t;


typedef struct texture_s
{
	char		name[16];
	unsigned	width, height;
	struct gltexture_s *gl_texture;
	int			anim_total;				// total tenths in sequence ( 0 = no)
	int			anim_min, anim_max;		// time for this frame min <=time< max
	struct texture_s *anim_next;		// in the animation sequence
	struct texture_s *alternate_anims;	// bmodels in frmae 1 use these

	// current animation
	struct texture_s *anim;

	// pointer to the data
	byte *data;

	// store the texture number for volume carving
	int num;
} texture_t;

// base surface types
#define	SURF_PLANEBACK		2
#define	SURF_DRAWSKY		4
#define SURF_DRAWSPRITE		8
#define SURF_DRAWTURB		16
#define SURF_DRAWTILED		32
#define SURF_DRAWBACKGROUND	64
#define SURF_UNDERWATER		128

// extra water surface types
#define SURF_DRAWLAVA		256
#define SURF_DRAWSLIME		512
#define SURF_DRAWTELE		1024
#define SURF_DRAWWATER		2048


// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct
{
	unsigned short	v[2];
	unsigned int	cachededgeoffset;
} medge_t;

typedef struct
{
	float		vecs[2][4];
	float		mipadjust;
	texture_t	*texture;
	int			flags;
} mtexinfo_t;


// define our customvertex structure...
// VERY Direct3D this, eh?
typedef struct polyverts_s
{
	float v[3];		// 0/1/2 - regular verts
	float tv[3];	// 3/4/5 - transformed verts
	float st[2];	// 6/7 - texture
	float lm[2];	// 8/9 - lightmap
	float st64[2];	// 10/11 - 64 x 64 version for detail/caustic/etc to keep consistent scale
	float warp[2];	// 12/13 - caustics warp
	float rgba[4];	// 14/15/16/17 - rgb colour for fogging
	float cm[3];	// 18/19/20 - normalization cubemap texcoords
} polyverts_t;

// we could put this on the hunk, but lets just get lazy instead
// double the size to allow for water subdivision
extern polyverts_t	surfvertlist[MAX_MAP_VERTS * 2];
extern float		*surfarray;
extern int			allocsurfverts;

// special vertex type for water
typedef struct waterverts_s
{
	float v[3];			// regular verts (no transforms needed here)
	float warp1[2];		// texture s and t for the first warp in TMU 0
	float warp2[2];		// texture s and t for the second warp in TMU 1
	float rgba[4];		// rgba colour for various hacks
} waterverts_t;

extern waterverts_t watervertlist[MAX_MAP_VERTS * 2];
extern float		*waterarray;
extern int			allocwaterverts;

typedef struct glpoly_s
{
	struct	glpoly_s	*next;
	struct	glpoly_s	*chain;
	int					numverts;
	int					flags;			// for SURF_UNDERWATER
	polyverts_t			*verts;
	waterverts_t		*water;
	int					vertindex;		// index into vertex array
	int					waterindex;		// index into water array

	vec3_t				midpoint;
	float				fxofs;
} glpoly_t;

// using dynamic allocation for lightmaps allows us to dramatically increase the number of available lightmaps
#define SINTABLE_SIZE	1024
#define LIGHTMAP_BYTES	4
#define	MAX_LIGHTMAPS	1024
#define	BLOCK_WIDTH		128
#define	BLOCK_HEIGHT	128

// not sure if these are needed anymore.  They're a legacy from a much older system I had in place a while back...
#define LM_FULLDARK		0
#define LM_FULLBRIGHT	255

typedef struct lightmapinfo_s
{
	int			gltexturenumber;
	int			staintexturenumber;
	qboolean	modified;
	int			allocated[BLOCK_WIDTH];
	byte		texels[LIGHTMAP_BYTES * BLOCK_WIDTH * BLOCK_HEIGHT];
	int			num;

	// stainmap extras
	byte		smtexels[LIGHTMAP_BYTES * BLOCK_WIDTH * BLOCK_HEIGHT];

	// true if the stainmap needs to be updated
	qboolean	stained;

	// true if it has ever been stained
	qboolean	stainactive;
} lightmapinfo_t;


typedef struct msurface_s
{
	int			visframe;		// should be drawn when node is crossed

	mplane_t	*plane;
	int			flags;

	int			firstedge;	// look up in model->surfedges[], negative numbers
	int			numedges;	// are backwards edges
	
	short		texturemins[2];
	short		extents[2];

	int			light_s, light_t;	// gl lightmap coordinates

	glpoly_t	*polys;				// multiple if warped
	byte		warpindex[2];		// warp texcoords

	struct	msurface_s	*texturechain;

	mtexinfo_t	*texinfo;

	// lighting info
	int			dlightframe;

	// dlights that might hit this surf
	int			dlightbits;

	// dlights that actually did hit this surf
	int			activedlightbits;

	// store a direct pointer to the surf lightmap
	lightmapinfo_t *lightmap;

	byte		styles[MAXLIGHTMAPS];
	int			cached_light[MAXLIGHTMAPS];	// values currently used in lightmap
	byte		*samples;		// [numstyles*surfsize]

	// lighting constants previously generated at runtime
	int			smax;
	int			tmax;
	int			base;

	// true if hit by a dlight
	qboolean	cached_dlight;

	// master drawlist
	struct msurface_s *drawlist;

	// world volume carving
	int			volume;
	texture_t	*volumetexture;

	int			surfnum;

	// true if visible for lava.  this hack doesn;t take frustum culling into account, as (most likely)
	// the lava surf will be culled but the smoke should still be visible.
	qboolean	lavavishack;

	// stainmaps
	unsigned short *stainsamples;

	// point checking
	vec3_t		mins;
	vec3_t		maxs;
} msurface_t;


typedef struct mnode_s
{
// common with leaf
	int			contents;		// 0, to differentiate from leafs
	int			visframe;		// node needs to be traversed if current
	
	float		minmaxs[6];		// for bounding box culling

	struct mnode_s	*parent;

// node specific
	mplane_t	*plane;
	struct mnode_s	*children[2];	

	unsigned short		firstsurface;
	unsigned short		numsurfaces;
} mnode_t;

typedef struct mleaf_s
{
// common with node
	int				contents;		// wil be a negative contents number
	int				visframe;		// node needs to be traversed if current

	vec3_t			mins;			// for bounding box culling
	vec3_t			maxs;			// for bounding box culling


	struct mnode_s	*parent;

// leaf specific
	byte			*decompressed_vis;
	efrag_t			*efrags;

	msurface_t		**firstmarksurface;
	int				nummarksurfaces;
	int				key;			// BSP sequence number for leaf's contents
	byte			ambient_sound_level[NUM_AMBIENTS];

	int				volume;
	texture_t		*volumetexture;
} mleaf_t;

// !!! if this is changed, it must be changed in asm_i386.h too !!!
typedef struct
{
	dclipnode_t	*clipnodes;
	mplane_t	*planes;
	int			firstclipnode;
	int			lastclipnode;
	vec3_t		clip_mins;
	vec3_t		clip_maxs;
} hull_t;

/*
==============================================================================

SPRITE MODELS

==============================================================================
*/


// moved from gl_rpart.c cos sprites need it too...
typedef enum
{
	P_POINT,
	P_SMOKE,
	P_BLOOD1,
	P_BLOOD2,
	P_BLOOD3,
	P_BLOOD4,
	P_SPARK,
	P_GUNSHOT,
	P_BLAST,
	P_FLARE1,
	P_FLARE2,
	P_BUBBLE,
	P_RAIN,
	P_SMOKE_NOSCALE,
	P_SMOKE_BLACK,
	P_SMOKE_BLACK_NOSCALE,
	P_NONE = 31
} pdrawstyle_t;


// FIXME: shorten these?
typedef struct mspriteframe_s
{
	int		width;
	int		height;
	float	up, down, left, right;
	struct gltexture_s	*gl_texture;
} mspriteframe_t;

typedef struct
{
	int				numframes;
	float			*intervals;
	mspriteframe_t	*frames[1];
} mspritegroup_t;

typedef struct
{
	spriteframetype_t	type;
	mspriteframe_t		*frameptr;
} mspriteframedesc_t;

typedef struct
{
	int					type;
	int					maxwidth;
	int					maxheight;
	int					numframes;
	pdrawstyle_t		replace;
	float				beamlength;		// remove?
	void				*cachespot;		// remove?
	mspriteframedesc_t	frames[1];
} msprite_t;


/*
==============================================================================

ALIAS MODELS

Alias models are position independent, so the cache manager can move them.
==============================================================================
*/

typedef struct maliasframedesc_s
{
	int					firstpose;
	int					numposes;
	float				interval;
	float				radius;
	vec3_t				bboxmin;
	vec3_t				bboxmax;
	int					frame;
	char				name[16];
} maliasframedesc_t;

typedef struct maliasgroupframedesc_s
{
	trivertx_t			bboxmin;
	trivertx_t			bboxmax;
	int					frame;
} maliasgroupframedesc_t;

typedef struct maliasgroup_s
{
	int						numframes;
	int						intervals;
	maliasgroupframedesc_t	frames[1];
} maliasgroup_t;

// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct mtriangle_s
{
	int					facesfront;
	int					vertindex[3];
} mtriangle_t;


#define	MAX_SKINS		1024
#define	MAXALIASVERTS	4096
#define	MAXALIASFRAMES	2048
#define	MAXALIASTRIS	8192

typedef struct aliashdr_s
{
	int					ident;
	int					version;
	vec3_t				scale;
	vec3_t				scale_origin;
	float				boundingradius;
	vec3_t				eyeposition;
	int					numskins;
	int					skinwidth;
	int					skinheight;
	int					numverts;
	int					numtris;
	int					numframes;
	synctype_t			synctype;
	int					flags;
	float				size;

	int					numposes;
	int					poseverts;
	int					posedata;	// numposes*poseverts trivert_t
	int					commands;	// gl command list with embedded s/t
	struct gltexture_s	*gl_texture[MAX_SKINS][4];
	int					texels[MAX_SKINS];	// only for player skins

	qboolean			glow;
	qboolean			torch;
	int					colour;
	qboolean			bolt;
	qboolean			powerup;
	qboolean			missile;
	qboolean			shambler;
	qboolean			flicker;
	qboolean			interpolate;
	qboolean			pulse;
	float				radius;
	float				translevel;
	maliasframedesc_t	frames[1];	// variable sized
} aliashdr_t;

extern	aliashdr_t	*pheader;
extern	stvert_t	stverts[MAXALIASVERTS];
extern	mtriangle_t	triangles[MAXALIASTRIS];
extern	trivertx_t	*poseverts[MAXALIASFRAMES];

//===================================================================

//
// Whole model
//
typedef enum {mod_brush, mod_sprite, mod_alias} modtype_t;

#define	EF_ROCKET	1			// leave a trail
#define	EF_GRENADE	2			// leave a trail
#define	EF_GIB		4			// leave a trail
#define	EF_ROTATE	8			// rotate (bonus items)
#define	EF_TRACER	16			// green split trail
#define	EF_ZOMGIB	32			// small blood trail
#define	EF_TRACER2	64			// orange split trail + rotate
#define	EF_TRACER3	128			// purple trail

typedef struct model_s
{
	char			name[MAX_QPATH];
	qboolean		loaded;			// bmodels and sprites don't cache normally

	modtype_t		type;
	int				numframes;
	int				maxframe;		// For warnings
	synctype_t		synctype;
	
	int				flags;

//
// volume occupied by the model graphics
//		
	vec3_t			mins, maxs;
	vec3_t			mids;
	float			radius;

//
// solid volume for clipping 
//
	qboolean		clipbox;
	vec3_t			clipmins, clipmaxs;

//
// brush model
//
	int				firstmodelsurface, nummodelsurfaces;

	int				numsubmodels;
	dmodel_t		*submodels;

	int				numplanes;
	mplane_t		*planes;

	int				numleafs;		// number of visible leafs, not counting 0
	mleaf_t			*leafs;

	int				numvertexes;
	mvertex_t		*vertexes;

	int				numedges;
	medge_t			*edges;

	int				numnodes;
	mnode_t			*nodes;

	int				numtexinfo;
	mtexinfo_t		*texinfo;

	int				numsurfaces;
	msurface_t		*surfaces;

	int				numsurfedges;
	int				*surfedges;

	int				numclipnodes;
	dclipnode_t		*clipnodes;

	int				nummarksurfaces;
	msurface_t		**marksurfaces;

	hull_t			hulls[MAX_MAP_HULLS];

	int				numtextures;
	texture_t		**textures;

	byte			*visdata;
	int				numvisdata;
	int				vissize;

	byte			*lightdata;
	int				numlightdata;

	char			*entities;
	int				numentities;

	// extra fields for lightpointing
	vec3_t			lightspot;
	vec3_t			pointcolour;
	mplane_t		*lightplane;

	// stainmaps
	unsigned short	*staindata;

	// additional model data
	void			*extradata;
} model_t;

// litfile header - darkplaces standard
typedef struct litheader_s
{
	char ident[4];
	int version;
} litheader_t;


//============================================================================

// 2d piccies
typedef struct
{
	struct gltexture_s *gl_texture;
	float	sl, tl, sh, th;
} glpic_t;

extern char	 loadname[32];

void	Mod_Init (void);
void	Mod_ClearAll (void);
model_t *Mod_ForName (char *name, qboolean crash);
mleaf_t *Mod_PointInLeaf (vec3_t point, model_t *model);
byte	*Mod_LeafPVS (mleaf_t *leaf, model_t *model);

#endif	// __MODEL__
