/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "gl_state.h"


// particle type definitions moved from glquake.h to here cos they're not used anywhere else!
// the "must be changed in d_ifacea.h too" bit is no longer valid either
typedef enum
{
	pt_static,
	pt_grav,
	pt_slowgrav,
	pt_smoketrail,
	pt_explode,
	pt_explode2,
	pt_explosion,
	pt_blob,
	pt_blob2,
	pt_kill,
	pt_lava,
	pt_explosmoke,
	pt_gunsmoke,
} ptype_t;

// by "driver" I assume they mean QC here...
typedef struct particle_s
{
	// driver-usable fields
	vec3_t				org;
	float				color;

	// drivers never touch the following fields
	struct particle_s	*next;
	vec3_t				vel;
	float				ramp;
	float				die;
	ptype_t				type;
	pdrawstyle_t		drawstyle;

	float				radius;
	float				alpha;

	// chaining
	struct particle_s	*pchain;
} particle_t;

// keep this high enough to kill framerate dependency as much as possible
#define NUM_TORCH_SMOKE_PARTICLES 1024
static particle_t R_SmokeParticles[NUM_TORCH_SMOKE_PARTICLES];

int		ramp1[8] = {0x6f, 0x6d, 0x6b, 0x69, 0x67, 0x65, 0x63, 0x61};
int		ramp2[8] = {0x6f, 0x6e, 0x6d, 0x6c, 0x6b, 0x6a, 0x68, 0x66};
int		ramp3[8] = {96, 98, 100, 6, 4, 3};

// massively increased so that a small handful of explosions don't max it out
#define MAX_PARTICLES			8192
static particle_t	particles[MAX_PARTICLES];
static particle_t	*active_particles, *free_particles;

typedef struct partdef_s
{
	// texcoords
	float	stmin[2];
	float	stmax[2];

	// sizing
	float	scale;
} partdef_t;

// this must at least equal the number of elements in the following enum.
// i've set 32 to give headroom...
#define NUM_PARTICLE_STYLES 32
static partdef_t partdef[NUM_PARTICLE_STYLES];

static float PRGBA[3][256];

/*
===============
R_InitTorchSmokeParticle
===============
*/
void R_InitTorchSmokeParticle (particle_t *p)
{
	// set the base particle state
	if (Q_rand () & 1)
	{
		p->color = 0 + (Q_rand () % 7);
	}
	else 
	{
		p->color = 96 + (Q_rand () % 7);
	}
	p->org[0] = -2 + (Q_rand () % 5);
	p->org[1] = -2 + (Q_rand () % 5);
	p->org[2] = -2 + (Q_rand () % 5);

	p->vel[0] = p->vel[1] = p->vel[2] = 0;

	p->drawstyle = P_SMOKE_BLACK_NOSCALE;
	p->radius = 3;
	p->alpha = 0.666f;
}

/*
===============
R_UpdateTorchSmokeParticle
===============
*/
void R_UpdateTorchSmokeParticle (particle_t *p, float updatetime)
{
	p->org[2] += p->vel[2] * updatetime;
	p->radius += (updatetime * 30);
	p->vel[2] += (updatetime * 45);
	p->alpha -= (updatetime / 2);
}

/*
===============
R_UpdateTorchSmokeParticles
===============
*/
void R_UpdateTorchSmokeParticles (void)
{
	int		i;
	float	lowestalpha = 666;
	int		lowestalphanum = 0;

	for (i = 0; i < NUM_TORCH_SMOKE_PARTICLES; i++)
	{
		R_UpdateTorchSmokeParticle (&R_SmokeParticles[i], r_frametime);

		// find the particle with the lowest alpha
		if (R_SmokeParticles[i].alpha < lowestalpha)
		{
			lowestalpha = R_SmokeParticles[i].alpha;
			lowestalphanum = i;
		}
	}

	// add a new particle to the list to replace the particle with the lowest alpha
	// done here to ensure that only one new particle is added each frame, otherwise the
	// smoke animation experiences some strange blips every now and then
	R_InitTorchSmokeParticle (&R_SmokeParticles[lowestalphanum]);
}

/*
===============
R_InitTorchSmokeParticles
===============
*/
void R_InitTorchSmokeParticles (void)
{
	int		i;
	int		j;
	float	updatetime;

	for (i = 0; i < NUM_TORCH_SMOKE_PARTICLES; i++)
	{
		R_InitTorchSmokeParticle (&R_SmokeParticles[i]);

		// run a fake update to simulate adding an extra particle each frame
		for (j = 0; j < i; j++)
		{
			updatetime = 6 + (Q_rand () % 9);
			updatetime /= 1000.0;

			R_UpdateTorchSmokeParticle (&R_SmokeParticles[j], updatetime);
		}
	}
}


/*
===============
R_InitParticles

particle number limiting is *so* 1996...
===============
*/
void R_InitParticles (void)
{
	int j;

	// hack the frametime for the fake updates
	float temp_frametime = r_frametime;

	R_InitTorchSmokeParticles ();

	for (j = 0; j < 512; j++)
	{
		// generate a fake frametime, keep it random for validity
		r_frametime = 6 + (Q_rand () % 9);
		r_frametime /= 1000.0;

		// run a fake update of the particles
		R_UpdateTorchSmokeParticles ();
	}

	// restore the original frametime
	r_frametime = temp_frametime;

	for (j = 0; j < 256; j++)
	{
		PRGBA[0][j] = 0.00392156862745 * ((byte *) &d_8to24table[j])[0];
		PRGBA[1][j] = 0.00392156862745 * ((byte *) &d_8to24table[j])[1];
		PRGBA[2][j] = 0.00392156862745 * ((byte *) &d_8to24table[j])[2];
	}

	partdef[P_POINT].stmin[0] = 0.25f;
	partdef[P_POINT].stmin[1] = 0;
	partdef[P_POINT].stmax[0] = 0.375f;
	partdef[P_POINT].stmax[1] = 0.25f;
	partdef[P_POINT].scale = 0.004f;

	partdef[P_SMOKE].stmin[0] = 0;
	partdef[P_SMOKE].stmin[1] = 0;
	partdef[P_SMOKE].stmax[0] = 0.25f;
	partdef[P_SMOKE].stmax[1] = 0.5f;
	partdef[P_SMOKE].scale = 0.001f;

	partdef[P_SMOKE_NOSCALE].stmin[0] = 0;
	partdef[P_SMOKE_NOSCALE].stmin[1] = 0;
	partdef[P_SMOKE_NOSCALE].stmax[0] = 0.25f;
	partdef[P_SMOKE_NOSCALE].stmax[1] = 0.5f;
	partdef[P_SMOKE_NOSCALE].scale = 0;

	partdef[P_SMOKE_BLACK].stmin[0] = 0;
	partdef[P_SMOKE_BLACK].stmin[1] = 0;
	partdef[P_SMOKE_BLACK].stmax[0] = 0.25f;
	partdef[P_SMOKE_BLACK].stmax[1] = 0.5f;
	partdef[P_SMOKE_BLACK].scale = 0.001f;

	partdef[P_SMOKE_BLACK_NOSCALE].stmin[0] = 0;
	partdef[P_SMOKE_BLACK_NOSCALE].stmin[1] = 0;
	partdef[P_SMOKE_BLACK_NOSCALE].stmax[0] = 0.25f;
	partdef[P_SMOKE_BLACK_NOSCALE].stmax[1] = 0.5f;
	partdef[P_SMOKE_BLACK_NOSCALE].scale = 0;

	for (j = 0; j < 4; j++)
	{
		partdef[P_BLOOD1 + j].stmin[0] = 0.375f + (0.125f * (float) j);
		partdef[P_BLOOD1 + j].stmin[1] = 0;
		partdef[P_BLOOD1 + j].stmax[0] = 0.5f + (0.125f * (float) j);
		partdef[P_BLOOD1 + j].stmax[1] = 0.25f;
		partdef[P_BLOOD1 + j].scale = 0.001f;
	}

	partdef[P_SPARK].stmin[0] = 0.875f;
	partdef[P_SPARK].stmin[1] = 0;
	partdef[P_SPARK].stmax[0] = 1;
	partdef[P_SPARK].stmax[1] = 0.25f;
	partdef[P_SPARK].scale = 0.001f;

	partdef[P_GUNSHOT].stmin[0] = 0.75f;
	partdef[P_GUNSHOT].stmin[1] = 0.5f;
	partdef[P_GUNSHOT].stmax[0] = 1;
	partdef[P_GUNSHOT].stmax[1] = 1;
	partdef[P_GUNSHOT].scale = 0.001f;

	partdef[P_BLAST].stmin[0] = 0;
	partdef[P_BLAST].stmin[1] = 0.5f;
	partdef[P_BLAST].stmax[0] = 0.25f;
	partdef[P_BLAST].stmax[1] = 1;
	partdef[P_BLAST].scale = 0.001f;

	partdef[P_FLARE1].stmin[0] = 0.25f;
	partdef[P_FLARE1].stmin[1] = 0.5f;
	partdef[P_FLARE1].stmax[0] = 0.5f;
	partdef[P_FLARE1].stmax[1] = 1;
	partdef[P_FLARE1].scale = 0.001f;

	partdef[P_FLARE2].stmin[0] = 0.5f;
	partdef[P_FLARE2].stmin[1] = 0.5f;
	partdef[P_FLARE2].stmax[0] = 0.75f;
	partdef[P_FLARE2].stmax[1] = 1;
	partdef[P_FLARE2].scale = 0.001f;

	partdef[P_BUBBLE].stmin[0] = 0.375f;
	partdef[P_BUBBLE].stmin[1] = 0.25f;
	partdef[P_BUBBLE].stmax[0] = 0.5f;
	partdef[P_BUBBLE].stmax[1] = 0.5f;
	partdef[P_BUBBLE].scale = 0;

	partdef[P_RAIN].stmin[0] = 0.5f;
	partdef[P_RAIN].stmin[1] = 0.25f;
	partdef[P_RAIN].stmax[0] = 0.625f;
	partdef[P_RAIN].stmax[1] = 0.5f;
	partdef[P_RAIN].scale = 0.001f;

	// null particle - just grab a bit of the texture that's all black
	partdef[P_NONE].stmin[0] = 0;
	partdef[P_NONE].stmin[1] = 0;
	partdef[P_NONE].stmax[0] = 0.025f;
	partdef[P_NONE].stmax[1] = 0.0125f;
	partdef[P_NONE].scale = 0;
}

/*
===============
R_EntityParticles
===============
*/

#define NUMVERTEXNORMALS	162
extern	float	r_avertexnormals[NUMVERTEXNORMALS][3];
vec3_t	avelocities[NUMVERTEXNORMALS];
float	beamlength = 16;
vec3_t	avelocity = {23, 7, 3};
float	partstep = 0.01f;
float	timescale = 0.01f;

void R_EntityParticles (entity_t *ent)
{
	int			count;
	int			i;
	particle_t	*p;
	float		angle;
	float		sr, sp, sy, cr, cp, cy;
	vec3_t		forward;
	float		dist;
	
	dist = 64;
	count = 50;

	if (!avelocities[0][0])
	{
		for (i=0 ; i<NUMVERTEXNORMALS*3 ; i++)
		avelocities[0][i] = (Q_rand()&255) * 0.01;
	}

	for (i=0 ; i<NUMVERTEXNORMALS ; i++)
	{
		angle = cl.time * avelocities[i][0];
		sy = sin(angle);
		cy = cos(angle);
		angle = cl.time * avelocities[i][1];
		sp = sin(angle);
		cp = cos(angle);
		angle = cl.time * avelocities[i][2];
		sr = sin(angle);
		cr = cos(angle);
	
		forward[0] = cp*cy;
		forward[1] = cp*sy;
		forward[2] = -sp;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = realtime + 0.01;
		p->color = 0x6f;
		p->type = pt_kill;
		
		p->org[0] = ent->origin[0] + r_avertexnormals[i][0]*dist + forward[0]*beamlength;			
		p->org[1] = ent->origin[1] + r_avertexnormals[i][1]*dist + forward[1]*beamlength;			
		p->org[2] = ent->origin[2] + r_avertexnormals[i][2]*dist + forward[2]*beamlength;	

		p->drawstyle = P_POINT;
		p->radius = 1;
		p->alpha = 1;
	}
}

/*
===============
R_ClearParticles
===============
*/
void R_ClearParticles (void)
{
	int		i;
	
	free_particles = &particles[0];
	active_particles = NULL;

	for (i = 0; i < MAX_PARTICLES; i++)
	{
		particles[i].next = &particles[i + 1];

		// stop undefined particles picking up the style from a previous defined one in this slot
		particles[i].drawstyle = P_POINT;
		particles[i].radius = 1;
	}
	particles[MAX_PARTICLES - 1].next = NULL;
}

/*
===============
R_ParticleExplosion
===============
*/
#define NUM_EXPLODE_PARTICLES 64
void R_ParticleExplosion (vec3_t org)
{
	int			i;
	particle_t	*p;

	for (i = 0; i < NUM_EXPLODE_PARTICLES; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = realtime + 15;
		p->color = ramp1[0];
		p->ramp = Q_rand () & 3;

		p->org[0] = org[0] + ((Q_rand () % 32) - 16);
		p->org[1] = org[1] + ((Q_rand () % 32) - 16);
		p->org[2] = org[2] + ((Q_rand () % 32) - 16);

		// explosion velocities don't get modified
		p->vel[0] = 0;
		p->vel[1] = 0;
		p->vel[2] = 0;

		p->radius = 64 + (Q_rand () % 96) - 48;
		p->drawstyle = P_BLAST;

		p->alpha = 1;
		p->type = pt_explosion;
	}
}


/*
===============
R_ParticleExplosion2
===============
*/
void R_ParticleExplosion2 (vec3_t org, int colorStart, int colorLength)
{
	int			i, j;
	particle_t	*p;
	int			colorMod = 0;

	for (i=0; i<512; i++)
	{
		if (!free_particles) return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = realtime + 0.3;
		p->color = colorStart + (colorMod % colorLength);
		colorMod++;

		p->type = pt_blob;

		for (j=0 ; j<3 ; j++)
		{
			p->org[j] = org[j] + ((Q_rand()%32)-16);
			p->vel[j] = (Q_rand()%512)-256;
		}
		p->drawstyle = P_POINT;
		p->radius = 1;
		p->alpha = 1;
	}
}

/*
===============
R_BlobExplosion
===============
*/
void R_BlobExplosion (vec3_t org)
{
	int			i, j;
	particle_t	*p;
	
	for (i=0 ; i<1024 ; i++)
	{
		if (!free_particles) return;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = realtime + 1 + (Q_rand()&8)*0.05;

		if (i & 1)
		{
			p->type = pt_blob;
			p->color = 66 + Q_rand()%6;

			for (j=0 ; j<3 ; j++)
			{
				p->org[j] = org[j] + ((Q_rand()%32)-16);
				p->vel[j] = (Q_rand()%512)-256;
			}
		}
		else
		{
			p->type = pt_blob2;
			p->color = 150 + Q_rand()%6;

			for (j=0 ; j<3 ; j++)
			{
				p->org[j] = org[j] + ((Q_rand()%32)-16);
				p->vel[j] = (Q_rand()%512)-256;
			}
		}
		p->drawstyle = P_POINT;
		p->radius = 1;
		p->alpha = 1;
	}
}

/*
===============
R_ParticleSprite
===============
*/
void R_ParticleSprite (int color, float radius, vec3_t origin, pdrawstyle_t ptype)
{
	particle_t	*p;

	if (!free_particles) return;

	p = free_particles;
	free_particles = p->next;
	p->next = active_particles;
	active_particles = p;

	// these are automatically destroyed when the particle is updated,
	// so give them enough time to ensure that the particle is drawn
	p->die = realtime + 666;

	p->color = (float)color;
	p->type = pt_kill;	// go into autodestruct mode

	p->drawstyle = ptype;
	p->radius = radius;
	p->alpha = 1;

	p->org[0] = origin[0];
	p->org[1] = origin[1];
	p->org[2] = origin[2];

	p->vel[0] = 0;
	p->vel[1] = 0;
	p->vel[2] = 0;
}


/*
===============
R_RunParticleEffect
===============
*/
void R_RunParticleEffect (vec3_t org, vec3_t dir, int color, int count)
{
	int			i, j;
	particle_t	*p;

	for (i = 0; i < count; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = realtime + 0.1 * (Q_rand () % 5);
		p->color = (color & ~7) + (Q_rand () & 7);
		p->type = pt_slowgrav;

		for (j = 0; j < 3; j++)
		{
			p->org[j] = org[j] + ((Q_rand () & 15) - 8);
			p->vel[j] = dir[j] * 15;
		}
		p->drawstyle = P_POINT;
		p->radius = 1;
		p->alpha = 1;
	}
}

/*
===============
R_RunGunshotEffect
===============
*/
void R_RunGunshotEffect (vec3_t org, int count)
{
	int			i;
	particle_t	*p;

	for (i = 0; i < count; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = realtime + 0.1 * (Q_rand () % 5);
		p->color = (107 & ~7) + (Q_rand () & 7);
		p->type = pt_static;

		p->org[0] = org[0] + ((Q_rand() & 3) - 2);
		p->org[1] = org[1] + ((Q_rand() & 3) - 2);
		p->org[2] = org[2] + ((Q_rand() & 3) - 2);

		p->vel[0] = 0;
		p->vel[1] = 0;
		p->vel[2] = 0;

		p->drawstyle = P_GUNSHOT;
		p->radius = 10;
		p->alpha = 1;
	}
}

/*
===============
R_SmokeEffect
===============
*/
void R_SmokeEffect (vec3_t org, int count, float InitialAlpha)
{
	int i;
	particle_t *p;

	for (i = 0; i < count; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		// set the die ludicrously high so that the smoke will always go through the full alpha ramp
		p->die = realtime + 666;

		// although the type is "gunsmoke" it can be used for any generic rising smoke type
		p->type = pt_gunsmoke;

		p->color = 0 + (Q_rand () % 7);

		p->org[0] = org[0] - 2 + (Q_rand () % 5);
		p->org[1] = org[1] - 2 + (Q_rand () % 5);
		p->org[2] = org[2] - 2 + (Q_rand () % 5);

		p->vel[0] = 0;
		p->vel[1] = 0;
		p->vel[2] = 0;

		p->drawstyle = P_SMOKE_BLACK;
		p->radius = 3;
		p->alpha = InitialAlpha;
	}
}

/*
===============
R_BloodParticleSplash
===============
*/
void R_BloodParticleSplash (vec3_t org, vec3_t dir, int color, int count)
{
	int			i;
	particle_t	*p;

	for (i = 0; i < count; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = realtime + 0.1 * (Q_rand () % 5);

		p->color = (int) ((color & ~7) + (Q_rand () & 7));

		p->type = pt_slowgrav;

		p->org[0] = org[0] + ((Q_rand () & 15) - 8);
		p->org[1] = org[1] + ((Q_rand () & 15) - 8);
		p->org[2] = org[2] + ((Q_rand () & 15) - 8);

		p->vel[0] = dir[0] * 15;
		p->vel[1] = dir[1] * 15;
		p->vel[2] = dir[2] * 15;

		p->drawstyle = P_BLOOD1 + (Q_rand () % 4);
		p->radius = 4 + (Q_rand () % 4);
		p->alpha = 1;
	}
}

/*
===============
R_HipnoticParticleField
===============
*/
void R_HipnoticParticleField (vec3_t org, vec3_t dir, int color, int count)
{
	int			i;
	particle_t	*p;

	for (i = 0; i < count * 5; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = realtime + 0.3 * (Q_rand () % 5);

		p->color = (int) ((color & ~7) + (Q_rand() & 7));

		p->type = pt_slowgrav;

		p->org[0] = org[0] + ((Q_rand () & 15) - 8);
		p->org[1] = org[1] + ((Q_rand () & 15) - 8);
		p->org[2] = org[2] + ((Q_rand () & 15) - 8);

		p->vel[0] = dir[0] * 15;
		p->vel[1] = dir[1] * 15;
		p->vel[2] = dir[2] * 15;

		p->drawstyle = P_SPARK;
		p->radius = 1;
		p->alpha = 1;
	}
}


/*
===============
R_ParseParticleEffect

Parse an effect out of the server message
===============
*/
void R_ParseParticleEffect (void)
{
	vec3_t		org, dir;
	int			i, count, color;

	for (i = 0; i < 3; i++) org[i] = MSG_ReadCoord ();
	for (i = 0; i < 3; i++) dir[i] = MSG_ReadChar () * (1.0 / 16);

	count = MSG_ReadByte ();
	color = MSG_ReadByte ();

	if (hipnotic && count <= 4)
	{
		// particle field
		R_HipnoticParticleField (org, dir, color, count);
	}
	else if (count == 255)
	{
		// always an explosion
		R_ParticleExplosion (org);
	}
	else if (color == 73 || color == 225)
	{
		// blood splashes
		R_BloodParticleSplash (org, dir, color, count);
	}
	else
	{
		// generic particle effect type thing (should be very few of these...)
		R_RunParticleEffect (org, dir, color, count);
	}
}


/*
===============
R_LavaSplash
===============
*/
void R_LavaSplash (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;

	for (i =- 16; i < 16; i++)
	{
		for (j =- 16; j < 16; j++)
		{
			for (k = 0; k < 1; k++)
			{
				if (!free_particles) return;
				
				p = free_particles;
				free_particles = p->next;
				p->next = active_particles;
				active_particles = p;
				
				p->die = realtime + 2 + (Q_rand () & 31) * 0.02;
				p->color = 224 + (Q_rand () & 7);
				p->type = pt_slowgrav;
				
				p->drawstyle = P_POINT;
				p->radius = 5 + (Q_rand () & 7);
				p->alpha = 1;
				
				dir[0] = j * 8 + (Q_rand () & 7);
				dir[1] = i * 8 + (Q_rand () & 7);
				dir[2] = 256;
				
				p->org[0] = org[0] + dir[0];
				p->org[1] = org[1] + dir[1];
				p->org[2] = org[2] + (Q_rand() & 63);
				
				NormalizeFast (dir);
				vel = 50 + (rand () & 63);
				VectorScale (dir, vel, p->vel);
			}
		}
	}

	for (i = -16; i <= 16; i += 2)
	{
		for (j = -16; j <= 16; j += 2)
		{
			for (k = 0; k < 1; k++)
			{
				if (!free_particles) return;

				p = free_particles;
				free_particles = p->next;
				p->next = active_particles;
				active_particles = p;

				p->die = realtime + 5 + (Q_rand () & 63) * 0.02;

				if (i > -15 && i < 15 && j > -15 && j < 15)
					p->color = 96 + (Q_rand () % 7);
				else 
					p->color = 0 + (Q_rand () % 7);

				p->type = pt_lava;

				dir[0] = j*8 + (Q_rand()&7);
				dir[1] = i*8 + (Q_rand()&7);
				dir[2] = 256;

				p->org[0] = org[0] + dir[0];
				p->org[1] = org[1] + dir[1];
				p->org[2] = org[2] + (Q_rand () & 63);

				NormalizeFast (dir);
				vel = 50 + (Q_rand () & 63);
				VectorScale (dir, vel, p->vel);

				p->drawstyle = P_SMOKE;
				p->radius = 20 + (Q_rand () % 20);
				p->alpha = 0.666f;
			}
		}
	}
}

/*
===============
R_SmokeSplash
===============
*/
void R_SmokeSplash (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;
	
	for (i = -3; i < 4; i++)
	{
		for (j = -3; j < 4; j++)
		{
			for (k = -3; k < 4; k++)
			{
				if (!free_particles) return;

				p = free_particles;
				free_particles = p->next;
				p->next = active_particles;
				active_particles = p;

				p->die = realtime + 1 + (Q_rand () & 31) * 0.02;

				if (Q_rand () & 1)
				{
					p->color = 0 + (Q_rand () % 7);
					p->alpha = 0.666f;
				}
				else
				{
					p->color = 96 + (Q_rand () % 7);
					p->alpha = 0.666f;
				}
				p->type = pt_explosmoke;

				// check positioning
				dir[0] = j * 12 + (Q_rand () & 7);
				dir[1] = i * 12 + (Q_rand () & 7);
				dir[2] = k * 12 + (Q_rand () & 7);

				p->org[0] = org[0] + dir[0];
				p->org[1] = org[1] + dir[1];
				p->org[2] = org[2] + dir[2];

				NormalizeFast (dir);
				vel = 12 + (Q_rand () & 15);

				p->vel[0] = dir[0] * vel;
				p->vel[1] = dir[1] * vel;
				p->vel[2] = dir[2] * vel;

				p->drawstyle = P_SMOKE_BLACK;
				p->radius = 20 + (Q_rand () % 20);
			}
		}
	}
}

/*
===============
R_TeleportSplash
===============
*/
void R_TeleportSplash (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;

	for (i = -16; i < 17; i += 4)
	{
		for (j = -16; j < 17; j += 4)
		{
			for (k = -32; k < 33; k += 4)
			{
				if (!free_particles) return;

				p = free_particles;
				free_particles = p->next;
				p->next = active_particles;
				active_particles = p;
		
				p->die = realtime + 0.5 + (Q_rand()&7) * 0.02;
				p->color = 7 + (Q_rand()&7);
				p->type = pt_slowgrav;
				
				dir[0] = j*8;
				dir[1] = i*8;
				dir[2] = k*8;
	
				p->org[0] = org[0] + i + (Q_rand()&3);
				p->org[1] = org[1] + j + (Q_rand()&3);
				p->org[2] = org[2] + k + (Q_rand()&3);
	
				NormalizeFast (dir);						
				vel = 50 + (Q_rand()&63);
				VectorScale (dir, vel, p->vel);

				switch (Q_rand () % 6)
				{
				case 1:
					p->drawstyle = P_FLARE1;
					p->radius = 1 + (Q_rand () % 8);
					break;

				case 2:
					p->drawstyle = P_FLARE2;
					p->radius = 1 + (Q_rand () % 8);
					break;

				default:
					p->drawstyle = P_SPARK;
					p->radius = 1 + (Q_rand () % 3);
					break;
				}

				p->alpha = 1;
			}
		}
	}
}

/*
===============
R_RocketTrail
===============
*/
void R_RocketTrail (vec3_t start, vec3_t end, int type)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	int			dec;
	static int	tracercount;

	VectorSubtract (end, start, vec);
	len = Normalize (vec);

	if (type < 128)
	{
		dec = 2;
	}
	else
	{
		dec = 1;
		type -= 128;
	}

	if (type < 2) 
	{
		dec = 1;
	}

	while (len > 0)
	{
		len -= dec;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;
		
		VectorCopy (vec3_origin, p->vel);
		p->die = realtime + 2;
		p->alpha = 1;

		switch (type)
		{
		case 0:	// rocket trail
			p->ramp = (Q_rand () & 3);
			p->color = ramp3[(int)p->ramp];
			p->type = pt_smoketrail;

			for (j = 0; j < 3; j++)
			{
				p->org[j] = start[j] + ((Q_rand()%6)-3);
			}
			p->drawstyle = P_SMOKE_BLACK;
			p->radius = 2;
			p->alpha = 0;
			break;

		case 1:	// smoke smoke
			p->ramp = (Q_rand() & 3) + 2;
			p->color = ramp3[(int)p->ramp];
			p->type = pt_smoketrail;

			for (j = 0; j < 3; j++)
			{
				p->org[j] = start[j] + ((Q_rand()%6)-3);
			}
			p->drawstyle = P_SMOKE_BLACK;
			p->radius = 2;
			p->alpha = 0;
			break;

		case 2:	// blood
			p->type = pt_grav;
			p->color = 67 + (Q_rand () & 3);

			for (j = 0; j < 3; j++)
			{
				p->org[j] = start[j] + ((Q_rand () % 6) - 3);
			}
			p->drawstyle = P_BLOOD1 + (Q_rand () % 4);
			p->radius = 4 + (Q_rand () % 4);

			len -= 2;
			break;

		case 3:
		case 5:	// tracer
			p->die = realtime + 0.5;
			p->type = pt_static;

			if (type == 3)
			{
				p->color = 52 + ((tracercount&4)<<1);
			}
			else
			{
				p->color = 230 + ((tracercount&4)<<1);
			}
			tracercount++;

			VectorCopy (start, p->org);

			if (tracercount & 1)
			{
				p->vel[0] = 30*vec[1];
				p->vel[1] = 30*-vec[0];
			}
			else
			{
				p->vel[0] = 30*-vec[1];
				p->vel[1] = 30*vec[0];
			}
			p->drawstyle = P_POINT;
			p->radius = 1;
			break;

		case 4:	// slight blood
			p->type = pt_grav;
			p->color = 67 + (Q_rand () & 3);

			for (j = 0; j < 3; j++)
			{
				p->org[j] = start[j] + ((Q_rand () % 6) - 3);
			}
			len -= 5;

			p->drawstyle = P_BLOOD1 + (Q_rand () % 4);
			p->radius = 4 + (Q_rand () % 4);
			break;

		case 6:	// voor trail
			p->color = 9*16 + 8 + (Q_rand()&3);
			p->type = pt_static;
			p->die = realtime + 0.3;
			for (j=0 ; j<3 ; j++)
			{
				p->org[j] = start[j] + ((Q_rand()&15)-8);
			}
			p->drawstyle = P_POINT;
			p->radius = 1;
			break;
		}
		VectorAdd (start, vec, start);
	}
}


// vertex array struct for particles
typedef struct partverts_s
{
	float v[3];
	float st[2];
	float rgba[4];
} partverts_t;

#define MAX_PARTVERTS 1024

// draw in quad batches
static partverts_t partverts[MAX_PARTVERTS];

// particle chains
static particle_t *pchain[2];

// particle state changes
static qboolean pstatechange;

// particle origin
static vec3_t pUp, pRight, pOrg;

/*
===============
R_KillParticles

linked list management
===============
*/
void R_KillParticles (void)
{
	particle_t *p, *kill;
	int pcount = 0;

	// jump to the first active particle in the list
	while (1)
	{
		kill = active_particles;

		if (kill && kill->die < realtime)
		{
			active_particles = kill->next;
			kill->next = free_particles;
			free_particles = kill;

			continue;
		}
		break;
	}

	// init the chains
	pchain[0] = NULL;
	pchain[1] = NULL;

	// go through the list killing as we go
	// if active_particles == NULL the for loop will handle it admirably
	for (p = active_particles; p; p = p->next)
	{
		// jump to the next active particle in the list
		while (1)
		{
			kill = p->next;

			if (kill && kill->die < realtime)
			{
				p->next = kill->next;
				kill->next = free_particles;
				free_particles = kill;

				continue;
			}
			break;
		}
		pcount++;

		// build particle chains because there are 2 seperate blendfuncs we need to draw here
		switch (p->drawstyle)
		{
		// handle recognised drawstyles - blend 1
		case P_POINT:
		case P_BLOOD1:
		case P_BLOOD2:
		case P_BLOOD3:
		case P_BLOOD4:
		case P_SMOKE_BLACK:
		case P_SMOKE_BLACK_NOSCALE:
			p->pchain = pchain[0];
			pchain[0] = p;
			break;

		// handle recognised drawstyles - blend 2
		case P_SMOKE:
		case P_SMOKE_NOSCALE:
		case P_SPARK:
		case P_GUNSHOT:
		case P_BLAST:
		case P_FLARE1:
		case P_FLARE2:
		case P_BUBBLE:
		case P_RAIN:
		case P_NONE:
			p->pchain = pchain[1];
			pchain[1] = p;
			break;

		// fix up any unset or invalid drawstyles
		default:
			p->alpha = 1;
			p->radius = 1;
			p->drawstyle = P_POINT;
			p->pchain = pchain[0];
			pchain[0] = p;
			break;
		}
	}
}

/*
===============
R_UpdateParticles

linked list management
===============
*/
void R_UpdateParticles (void)
{
	particle_t		*p;
	float			time2, time3;
	float			time1;
	float			time0;
	float			halftime;
	float			explotime;
	float			tenthtime;
	float			dvel;
	float			grav;
	int				i;

	time3 = r_frametime * 15;
	time2 = r_frametime * 10;
	time1 = r_frametime * 5;
	time0 = r_frametime * 2;
	halftime = r_frametime / 2.0f;
	explotime = r_frametime / 3.5f;
	tenthtime = r_frametime / 10.0f;
	grav = r_frametime * sv_gravity.value * 0.05;
	dvel = 4 * r_frametime;

	for (p = active_particles; p; p = p->next)
	{
		// update the particle
		p->org[0] += p->vel[0] * r_frametime;
		p->org[1] += p->vel[1] * r_frametime;
		p->org[2] += p->vel[2] * r_frametime;

		switch (p->type)
		{
		case pt_smoketrail:
			// simulate brownian motion
			p->org[0] += (float) ((Q_rand () & 2) - 1) / 2;
			p->org[1] += (float) ((Q_rand () & 2) - 1) / 2;
			p->org[2] += (float) ((Q_rand () & 2) - 1) / 2;

			p->ramp += time1;
			p->radius += r_frametime * (Q_rand () % 30);

			if (p->ramp >= 6)
				p->die = -1;
			else
				p->color = ramp3[(int) p->ramp];

			p->alpha += time0;

			if (p->alpha > 0.666f) 
			{
				p->alpha = 0.666f;
			}
			p->vel[2] += grav;
			break;

		case pt_explosion:
			// the explosion effect should last the same time as the explosion dlight
			p->ramp += time2;

			if (p->ramp >= 8)
				p->die = -1;
			else
				p->color = ramp1[(int) p->ramp];

			break;

		case pt_explode:
			p->ramp += time2;
			if (p->ramp >=8)
				p->die = -1;
			else
				p->color = ramp1[(int)p->ramp];
			for (i=0 ; i<3 ; i++)
				p->vel[i] += p->vel[i]*dvel;
			p->vel[2] -= grav;
			break;

		case pt_explode2:
			p->ramp += time3;
			if (p->ramp >=8)
				p->die = -1;
			else
				p->color = ramp2[(int)p->ramp];
			for (i=0 ; i<3 ; i++)
				p->vel[i] -= p->vel[i]*r_frametime;
			p->vel[2] -= grav;
			break;

		case pt_blob:
			for (i=0 ; i<3 ; i++)
				p->vel[i] += p->vel[i]*dvel;
			p->vel[2] -= grav;
			break;

		case pt_blob2:
			for (i=0 ; i<2 ; i++)
				p->vel[i] -= p->vel[i]*dvel;
			p->vel[2] -= grav;
			break;

		case pt_grav:
		case pt_slowgrav:
			p->vel[2] -= grav;
			break;

		case pt_kill:
			// these are sprites that get redrawn every frame, so kill them now
			p->die = -1;
			break;

		case pt_explosmoke:
			// simulate brownian motion
			p->org[0] += (float) ((Q_rand () & 2) - 1) / 2;
			p->org[1] += (float) ((Q_rand () & 2) - 1) / 2;
			p->org[2] += (float) ((Q_rand () & 2) - 1) / 2;

			p->radius += r_frametime * (Q_rand () % 30);
			p->alpha -= explotime;
			p->die += 1;

			if (p->alpha <= 0.0f) p->die = -1;
			break;

		case pt_lava:
			// simulate brownian motion
			p->org[0] += (float) ((Q_rand () & 2) - 1) / 2;
			p->org[1] += (float) ((Q_rand () & 2) - 1) / 2;
			p->org[2] += (float) ((Q_rand () & 2) - 1) / 2;

			p->radius += r_frametime * (Q_rand () % 30);

			p->vel[2] -= grav;

			p->alpha -= tenthtime;
			p->die += 1;

			if (p->alpha <= 0.0f) p->die = -1;
			break;

		case pt_gunsmoke:
			// gunsmoke grows and fades
			p->radius += time2 * 3;
			p->vel[2] += grav;
			p->alpha -= r_frametime;
			if (p->alpha < 0) p->die = -1;
			break;

		default:
			// unset or invalid falls through to static
		case pt_static:
			break;
		}
	}
}


/*
===============
R_ExecStateChangeForParticles

keep track of whether we have had a state change
execute the required state change for particle drawing
===============
*/
void R_ExecStateChangeForParticles (void)
{
	float *pv = (float *) partverts;

	// state change required for particles
	glDepthMask (GL_FALSE);
	glEnable (GL_BLEND);

	glBindTexture (GL_TEXTURE_2D, particletexture.gltexnum);
	glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glEnableClientState (GL_VERTEX_ARRAY);
	glClientActiveTexture (GL_TEXTURE0);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glEnableClientState (GL_COLOR_ARRAY);

	glVertexPointer (3, GL_FLOAT, sizeof (partverts_t), &pv[0]);
	glTexCoordPointer (2, GL_FLOAT, sizeof (partverts_t), &pv[3]);
	glColorPointer (4, GL_FLOAT, sizeof (partverts_t), &pv[5]);

	// record that we've had a state change
	pstatechange = true;
}

/*
===============
R_DrawParticleChain
===============
*/
void R_DrawParticleChain (particle_t *p, GLenum MH_GL_BLENDFUNC_SRC_FUNC, GLenum MH_GL_BLENDFUNC_DST_FUNC)
{
	partdef_t		*pd;
	partverts_t		*pv = partverts;
	float			scale;
	int				pbatch = 0;
	float			dist[3];
	float			rad;

	// set the required state if needed
	if (!pstatechange) 
	{
		R_ExecStateChangeForParticles ();
	}

	// set the correct blendfgunc for this chain
	glBlendFunc (MH_GL_BLENDFUNC_SRC_FUNC, MH_GL_BLENDFUNC_DST_FUNC);

	for (/**/; p; p = p->pchain)
	{
		// fillrate damage limitation
		// these 2 types are the big fillrate damagers up front, so do some magic stuff to them
		if (p->type == pt_explosmoke || p->drawstyle == P_BLAST)
		{
			// test for bbox culling around the particle.
			// this can help *a lot* when chewbacca goes to town on the fillrate.
			// just use the origin.  we could build a proper bbox by adding and subtracting
			// rad to get the maxs and mins, but in practice it doesn't matter.
			if (R_CullBox (p->org, p->org)) continue;

			// passed the culling test, so now do the distance test
			VectorSubtract (p->org, r_origin, dist);

			// multiply out into dist[0]
			dist[0] = dist[0] * dist[0] + dist[1] * dist[1] + dist[2] * dist[2];

			// reduce by a factor of 0.5 to 1 if closer than 256 units
			// this prevents it becoming so small that it's weedy, but still reduces the damage quite a bit
			// it's an agreeable enough compromise in other words...
			if (dist[0] < 1024) continue;

			if (dist[0] < 65536)
			{
				// scale the radius down
				// dist[0] of 65536 == 256 units (sqrt (65536))
				rad = (Q_sqrt (dist[0]) * 0.001953125 + 0.5) * p->radius;
			}
			else
			{
				// far enough away so keep the radius
				rad = p->radius;
			}
		}
		else
		{
			rad = p->radius;
		}

		// calculate the scale
		scale = (p->org[0] - r_origin[0]) * vpn[0] + (p->org[1] - r_origin[1]) * vpn[1] + (p->org[2] - r_origin[2]) * vpn[2];

		// get the current particle definition
		pd = &partdef[p->drawstyle];

		if (scale < 10)	continue;

		if (scale < 20)
		{
			// normal size
			scale = 1;
		}
		else
		{
			// hack a scale up to keep particles from disapearing
			scale = 1 + scale * pd->scale;
		}

		// save some calcs later on
		rad *= scale;

		// calculate the scaled vertex offset factors
		VectorScale (vup, rad, pUp);
		VectorScale (vright, rad, pRight);

		// these particles will be off-center, so correct that by calcing a new origin
		// to offset the verts from, which is half the offset away in each direction
		pOrg[0] = p->org[0] - (pRight[0] + pUp[0]) / 2;
		pOrg[1] = p->org[1] - (pRight[1] + pUp[1]) / 2;
		pOrg[2] = p->org[2] - (pRight[2] + pUp[2]) / 2;

		// reverse the winding order so we don't have to enable backface culling
		pv[0].v[0] = pOrg[0];
		pv[0].v[1] = pOrg[1];
		pv[0].v[2] = pOrg[2];

		pv[0].st[0] = pd->stmin[0];
		pv[0].st[1] = pd->stmin[1];

		pv[1].v[0] = pOrg[0] + pUp[0];
		pv[1].v[1] = pOrg[1] + pUp[1];
		pv[1].v[2] = pOrg[2] + pUp[2];

		pv[1].st[0] = pd->stmax[0];
		pv[1].st[1] = pd->stmin[1];

		pv[2].v[0] = pOrg[0] + pUp[0] + pRight[0];
		pv[2].v[1] = pOrg[1] + pUp[1] + pRight[1];
		pv[2].v[2] = pOrg[2] + pUp[2] + pRight[2];

		pv[2].st[0] = pd->stmax[0];
		pv[2].st[1] = pd->stmax[1];

		pv[3].v[0] = pOrg[0] + pRight[0];
		pv[3].v[1] = pOrg[1] + pRight[1];
		pv[3].v[2] = pOrg[2] + pRight[2];

		pv[3].st[0] = pd->stmin[0];
		pv[3].st[1] = pd->stmax[1];

		// colour is the same for all verts
		pv[0].rgba[0] = pv[1].rgba[0] = pv[2].rgba[0] = pv[3].rgba[0] = PRGBA[0][(int) p->color];
		pv[0].rgba[1] = pv[1].rgba[1] = pv[2].rgba[1] = pv[3].rgba[1] = PRGBA[1][(int) p->color];
		pv[0].rgba[2] = pv[1].rgba[2] = pv[2].rgba[2] = pv[3].rgba[2] = PRGBA[2][(int) p->color];
		pv[0].rgba[3] = pv[1].rgba[3] = pv[2].rgba[3] = pv[3].rgba[3] = p->alpha;

		pbatch += 4;

		if (pbatch >= MAX_PARTVERTS)
		{
			// draw arrays in quad batches.
			pbatch = 0;
			glDrawArrays (GL_QUADS, 0, MAX_PARTVERTS);
			pv = &partverts[pbatch];
		}
		else
		{
			// next batch
			pv = &partverts[pbatch];
		}
	}

	// draw anything left over
	if (pbatch) 
	{
		glDrawArrays (GL_QUADS, 0, pbatch);
	}
}

//===========================================================================================//
//									SMOKE EFFCTS											 //
//===========================================================================================//

#define LAVA_SMOKE_SCALE 12

// global counter for number of smoke particles
static int			smokecount = 0;
static partverts_t	*smokes;

/*
===============
R_DrawLavaSmokeEffect

crashes bastion of the underworld.
===============
*/
void R_DrawLavaSmokeEffect (vec3_t org, float ofs)
{
	int			i;
	int			base = abs (((int) ofs % 8));

	for (i = base; i < NUM_TORCH_SMOKE_PARTICLES; i += 16)
	{
		// this is just a container to keep things neat for the particle currently being drawn
		particle_t	p;
		partdef_t	*pd;

		// don't add anything with negative alpha
		// (allow ourselves to take some of the particles with negative alpha to get a better height range,
		// we'll add 0.25 to the alpha before drawing)
		if (R_SmokeParticles[i].alpha < -0.25) continue;

		// don't draw if the particle is completely under the lava
		if (org[2] + R_SmokeParticles[i].org[2] * (LAVA_SMOKE_SCALE / 4) + ofs < org[2] - 64) continue;

		// get the current particle definition
		pd = &partdef[R_SmokeParticles[i].drawstyle];

		p.org[0] = org[0] + R_SmokeParticles[i].org[0] * LAVA_SMOKE_SCALE;
		p.org[1] = org[1] + R_SmokeParticles[i].org[1] * LAVA_SMOKE_SCALE;
		p.org[2] = org[2] + R_SmokeParticles[i].org[2] * (LAVA_SMOKE_SCALE / 4) + ofs;

		// add 0.25 to alpha to compensate for the negative alpha of 0.25 we accepted above
		p.alpha = R_SmokeParticles[i].alpha + 0.25;

		// clamp the range
		if (p.alpha > 1) p.alpha = 1;
		if (p.alpha < 0) p.alpha = 0;

		// these particles will be off-center, so correct that by calcing a new origin
		// to offset the verts from, which is half the offset away in each direction
		pOrg[0] = p.org[0] - (pRight[0] + pUp[0]) / 2;
		pOrg[1] = p.org[1] - (pRight[1] + pUp[1]) / 2;
		pOrg[2] = p.org[2] - (pRight[2] + pUp[2]) / 2;

		// run the verts through our STSinTable to get some more warping on them
		smokes[0].v[0] = pOrg[0] + STSinTable[(int) ((pOrg[2] * 0.025 + realtime) * STMult) & STBitMask];
		smokes[0].v[1] = pOrg[1];
		smokes[0].v[2] = pOrg[2] + STSinTable[(int) ((pOrg[0] * 0.025 + realtime) * STMult) & STBitMask];
		
		smokes[0].rgba[0] = p.alpha * 1.5;
		smokes[0].rgba[1] = p.alpha * 0.7;
		smokes[0].rgba[2] = 0;
		smokes[0].rgba[3] = p.alpha * 3;

		smokes[0].st[0] = pd->stmin[0];
		smokes[0].st[1] = pd->stmin[1];

		smokes[1].v[0] = pOrg[0] + pUp[0] + STSinTable[(int) (((pOrg[2] + pUp[2]) * 0.025 + realtime) * STMult) & STBitMask];
		smokes[1].v[1] = pOrg[1] + pUp[1];
		smokes[1].v[2] = pOrg[2] + pUp[2] + STSinTable[(int) (((pOrg[0] + pUp[0]) * 0.025 + realtime) * STMult) & STBitMask];
		
		smokes[1].rgba[0] = p.alpha * 1.5;
		smokes[1].rgba[1] = p.alpha * 0.7;
		smokes[1].rgba[2] = 0;
		smokes[1].rgba[3] = p.alpha * 3;

		smokes[1].st[0] = pd->stmax[0];
		smokes[1].st[1] = pd->stmin[1];

		smokes[2].v[0] = pOrg[0] + pUp[0] + pRight[0] + STSinTable[(int) (((pOrg[2] + pUp[2] + pRight[2]) * 0.025 + realtime) * STMult) & STBitMask];
		smokes[2].v[1] = pOrg[1] + pUp[1] + pRight[1];
		smokes[2].v[2] = pOrg[2] + pUp[2] + pRight[2] + STSinTable[(int) (((pOrg[0] + pUp[0] + pRight[0]) * 0.025 + realtime) * STMult) & STBitMask];
		
		smokes[2].rgba[0] = p.alpha * 1.5;
		smokes[2].rgba[1] = p.alpha * 0.7;
		smokes[2].rgba[2] = 0;
		smokes[2].rgba[3] = p.alpha * 3;

		smokes[2].st[0] = pd->stmax[0];
		smokes[2].st[1] = pd->stmax[1];

		smokes[3].v[0] = pOrg[0] + pRight[0] + STSinTable[(int) (((pOrg[2] + pRight[2]) * 0.025 + realtime) * STMult) & STBitMask];
		smokes[3].v[1] = pOrg[1] + pRight[1];
		smokes[3].v[2] = pOrg[2] + pRight[2] + STSinTable[(int) (((pOrg[0] + pRight[0]) * 0.025 + realtime) * STMult) & STBitMask];
		
		smokes[3].rgba[0] = p.alpha * 1.5;
		smokes[3].rgba[1] = p.alpha * 0.7;
		smokes[3].rgba[2] = 0;
		smokes[3].rgba[3] = p.alpha * 3;

		smokes[3].st[0] = pd->stmin[0];
		smokes[3].st[1] = pd->stmax[1];

		smokecount += 4;

		if (smokecount >= MAX_PARTVERTS)
		{
			// draw arrays in quad batches.
			glDrawArrays (GL_QUADS, 0, smokecount);

			// reinit
			smokecount = 0;
			smokes = &partverts[smokecount];
		}
		else
		{
			// next batch
			smokes = &partverts[smokecount];
		}
	}
}

/*
===============
R_DrawTorchSmokeEffect

crashes bastion of the underworld.
===============
*/
void R_DrawTorchSmokeEffect (vec3_t org, float ofs)
{
	int			i;

	for (i = 0; i < NUM_TORCH_SMOKE_PARTICLES; i++)
	{
		// this is just a container to keep things neat for the particle currently being drawn
		particle_t	p;
		partdef_t	*pd;

		// don't add anything with negative alpha
		if (R_SmokeParticles[i].alpha < 0) continue;

		// get the current particle definition
		pd = &partdef[R_SmokeParticles[i].drawstyle];

		p.org[0] = org[0] + R_SmokeParticles[i].org[0];
		p.org[1] = org[1] + R_SmokeParticles[i].org[1];
		p.org[2] = org[2] + (R_SmokeParticles[i].org[2] * 1.25) + ofs;

		// calculate the scaled vertex offset factors
		VectorScale (vup, R_SmokeParticles[i].radius, pUp);
		VectorScale (vright, R_SmokeParticles[i].radius, pRight);

		// these particles will be off-center, so correct that by calcing a new origin
		// to offset the verts from, which is half the offset away in each direction
		pOrg[0] = p.org[0] - (pRight[0] + pUp[0]) / 2;
		pOrg[1] = p.org[1] - (pRight[1] + pUp[1]) / 2;
		pOrg[2] = p.org[2] - (pRight[2] + pUp[2]) / 2;

		smokes[0].v[0] = pOrg[0];
		smokes[0].v[1] = pOrg[1];
		smokes[0].v[2] = pOrg[2];

		smokes[0].rgba[0] = PRGBA[0][(int) R_SmokeParticles[i].color];
		smokes[0].rgba[1] = PRGBA[1][(int) R_SmokeParticles[i].color];
		smokes[0].rgba[2] = PRGBA[2][(int) R_SmokeParticles[i].color];
		smokes[0].rgba[3] = R_SmokeParticles[i].alpha;

		smokes[0].st[0] = pd->stmin[0];
		smokes[0].st[1] = pd->stmin[1];

		smokes[1].v[0] = pOrg[0] + pUp[0];
		smokes[1].v[1] = pOrg[1] + pUp[1];
		smokes[1].v[2] = pOrg[2] + pUp[2];

		smokes[1].rgba[0] = PRGBA[0][(int) R_SmokeParticles[i].color];
		smokes[1].rgba[1] = PRGBA[1][(int) R_SmokeParticles[i].color];
		smokes[1].rgba[2] = PRGBA[2][(int) R_SmokeParticles[i].color];
		smokes[1].rgba[3] = R_SmokeParticles[i].alpha;

		smokes[1].st[0] = pd->stmax[0];
		smokes[1].st[1] = pd->stmin[1];

		smokes[2].v[0] = pOrg[0] + pUp[0] + pRight[0];
		smokes[2].v[1] = pOrg[1] + pUp[1] + pRight[1];
		smokes[2].v[2] = pOrg[2] + pUp[2] + pRight[2];

		smokes[2].rgba[0] = PRGBA[0][(int) R_SmokeParticles[i].color];
		smokes[2].rgba[1] = PRGBA[1][(int) R_SmokeParticles[i].color];
		smokes[2].rgba[2] = PRGBA[2][(int) R_SmokeParticles[i].color];
		smokes[2].rgba[3] = R_SmokeParticles[i].alpha;

		smokes[2].st[0] = pd->stmax[0];
		smokes[2].st[1] = pd->stmax[1];

		smokes[3].v[0] = pOrg[0] + pRight[0];
		smokes[3].v[1] = pOrg[1] + pRight[1];
		smokes[3].v[2] = pOrg[2] + pRight[2];

		smokes[3].rgba[0] = PRGBA[0][(int) R_SmokeParticles[i].color];
		smokes[3].rgba[1] = PRGBA[1][(int) R_SmokeParticles[i].color];
		smokes[3].rgba[2] = PRGBA[2][(int) R_SmokeParticles[i].color];
		smokes[3].rgba[3] = R_SmokeParticles[i].alpha;

		smokes[3].st[0] = pd->stmin[0];
		smokes[3].st[1] = pd->stmax[1];

		smokecount += 4;

		if (smokecount >= MAX_PARTVERTS)
		{
			// draw arrays in quad batches.
			glDrawArrays (GL_QUADS, 0, smokecount);

			// reinit
			smokecount = 0;
			smokes = &partverts[smokecount];
		}
		else
		{
			// next batch
			smokes = &partverts[smokecount];
		}
	}
}

/*
===============
R_DrawLavaStencilSmoke
===============
*/
void R_DrawLavaStencilSmoke (msurface_t *surf, vec3_t org, float ofs)
{
	int			i;
	int			base = abs (((int) ofs % 8));
	vec3_t		sorg;
	float		scale[3];
	vec3_t		dist;

	for (i = base; i < NUM_TORCH_SMOKE_PARTICLES; i += 16)
	{
		// don't add anything with negative alpha
		// (allow ourselves to take some of the particles with negative alpha to get a better height range,
		// we'll add 0.25 to the alpha before drawing)
		if (R_SmokeParticles[i].alpha < -0.15) continue;

		// write directly into the translation matrix
		sorg[0] = org[0] + R_SmokeParticles[i].org[0] * LAVA_SMOKE_SCALE;
		sorg[1] = org[1] + R_SmokeParticles[i].org[1] * LAVA_SMOKE_SCALE;
		sorg[2] = org[2] + R_SmokeParticles[i].org[2] * (LAVA_SMOKE_SCALE / 4) + ofs;

		// run the verts through our STSinTable to get some more warping on them
		sorg[0] += STSinTable[(int) ((sorg[2] * 0.025 + realtime) * STMult) & STBitMask];
		sorg[2] += STSinTable[(int) ((sorg[0] * 0.025 + realtime) * STMult) & STBitMask];

		// bbox test to prevent wall poke-through
		if (!R_BoxContainsPoint (sorg, surf->mins, surf->maxs, 0)) continue;

		// distance test
		VectorSubtract (sorg, r_origin, dist);

		// multiply out into dist[0]
		dist[0] = dist[0] * dist[0] + dist[1] * dist[1] + dist[2] * dist[2];

		// too close
		if (dist[0] < 16384) continue;

		// modulate the scale by a sine wave in 3 dimensions
		scale[0] = 0.0008 * STSinTable[(int) ((R_SmokeParticles[i].org[2] * 0.25 + realtime) * STMult) & STBitMask];
		scale[1] = 0.0008 * STSinTable[(int) ((R_SmokeParticles[i].org[0] * 0.25 + realtime) * STMult) & STBitMask];
		scale[2] = 0.0008 * STSinTable[(int) ((R_SmokeParticles[i].org[1] * 0.25 + realtime) * STMult) & STBitMask];

		// in theory we could just load up the relevant matrix rather than doing push/pop, but
		// in practice it's not too reliable
		glPushMatrix ();
		glTranslatef (sorg[0], sorg[1], sorg[2]);
		glScalef (scale[0], scale[1], scale[2]);
		glCallList (spherelist2);
		glPopMatrix ();
		glFlush ();
	}
}

/*
===============
R_StencilHazeParticles
===============
*/
void R_StencilHazeParticles (void)
{
	int			i;
	msurface_t	*surf;
	glpoly_t	*p;

	// un-sex the render path
	glDisable (GL_TEXTURE_2D);

	// don't draw into the colour buffer
	glColorMask (0, 0, 0, 0);

	// take a quick stroll through the surfs
	// this doesn't really hurt things too much in terms of optimisation/etc.  we *could* build a lavachain
	// based on what's drawn and work from there, but that's not good cos then frustum culling would cause
	// smoke to drop out if the lava surf was culled... no really ideal solution.  we build a non-culled
	// chain, but that would involve extra checking in R_RecursiveWorldNode.  winning in one place means
	// losing in another whatever we do here...
	for (i = 0, surf = cl.worldmodel->surfaces; i < cl.worldmodel->numsurfaces; i++, surf++)
	{
		if (!(surf->flags & SURF_DRAWLAVA)) continue;
		if (!surf->lavavishack) continue;

		// ok, we have a lava surf which is visible
		for (p = surf->polys; p; p = p->next)
		{
			// draw a smoke effect type thing at the midpoint
			R_DrawLavaStencilSmoke (surf, p->midpoint, p->fxofs);
		}
	}

	// re-enable colour buffer
	glColorMask (1, 1, 1, 1);

	// some brain-dead drivers don;t restore from glColorMask properly, so...
	glColor4f (1, 1, 1, 1);

	glEnable (GL_TEXTURE_2D);
}


/*
===============
R_DrawSmokeParticles

Add in smoke effects.

Because there's going to be a shitload of these every frame, draw them seperately
from the main particle system
===============
*/
void R_DrawSmokeParticles (void)
{
	int			i;
	glpoly_t	*p;
	msurface_t	*surf;
	qboolean	pblendfunc = false;

	// set the initial verts and counter
	smokecount = 0;
	smokes = &partverts[smokecount];

	// take a quick stroll through the surfs
	// this doesn't really hurt things too much in terms of optimisation/etc.  we *could* build a lavachain
	// based on what's drawn and work from there, but that's not good cos then frustum culling would cause
	// smoke to drop out if the lava surf was culled... no really ideal solution.  we build a non-culled
	// chain, but that would involve extra checking in R_RecursiveWorldNode.  winning in one place means
	// losing in another whatever we do here...
	for (i = 0, surf = cl.worldmodel->surfaces; i < cl.worldmodel->numsurfaces; i++, surf++)
	{
		if (!(surf->flags & SURF_DRAWLAVA)) continue;
		if (!surf->lavavishack) continue;

		// ok, we have a lava surf which is visible
		for (p = surf->polys; p; p = p->next)
		{
			// set the required state if needed
			if (!pstatechange)
			{
				R_ExecStateChangeForParticles ();
			}

			// set the correct blendfunc for this chain
			if (!pblendfunc)
			{
				pblendfunc = true;

				glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

				// calculate the scaled vertex offset factors
				// only once for lava
				// anything higher than this and we start getting performance issues up close
				VectorScale (vup, 72, pUp);
				VectorScale (vright, 72, pRight);
			}

			// draw a smoke effect type thing at the midpoint
			R_DrawLavaSmokeEffect (p->midpoint, p->fxofs);
		}
	}

	// this can be exceeded when crossing the 2 batches, so deal with anything left over from
	// the first batch before moving to the second
	if (smokecount)
	{
		glDrawArrays (GL_QUADS, 0, smokecount);

		// reinit
		smokecount = 0;
		smokes = &partverts[smokecount];
	}

	// reset to false in case i ever decide to change the blendfunc below
	pblendfunc = false;

	// now do torch smoke
	for (i = 0; i < cl_numvisedicts; i++)
	{
		if (cl_visedicts[i]->model->type != mod_alias) continue;

		// torches only
		if (!((aliashdr_t *) (cl_visedicts[i]->model->extradata))->torch) continue;

		// set the required state if needed
		if (!pstatechange) 
		{
			R_ExecStateChangeForParticles ();
		}

		// set the correct blendfunc for this chain
		if (!pblendfunc)
		{
			glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			pblendfunc = true;
		}
		R_DrawTorchSmokeEffect (cl_visedicts[i]->origin, 4);
	}

	// draw anything left over
	if (smokecount) 
	{
		glDrawArrays (GL_QUADS, 0, smokecount);
	}

	// always required even if no particles are drawn
	R_UpdateTorchSmokeParticles ();
}

/*
===============
R_DrawParticles

Handles particle drawing.  Killing and updating are done seperately in order to make the drawing
operation more efficient.  Probably 6 of one, but hey...
===============
*/
void R_DrawParticles (void)
{
	float *pv = (float *) partverts;

	// kill old particles
	R_KillParticles ();

	// no state change
	pstatechange = false;

	// draw particle chains.  these are disabled in demo playback because they give a false lowering of
	// the times which can be misleading when benchmarking other effects.
	if (pchain[0]) 
	{
		R_DrawParticleChain (pchain[0], GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}

	if (pchain[1]) 
	{
		R_DrawParticleChain (pchain[1], GL_SRC_ALPHA, GL_ONE);
	}

	// add in smoke effects here too...
	R_DrawSmokeParticles ();

	if (pstatechange)
	{
		// reset state
		glDisableClientState (GL_VERTEX_ARRAY);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
		glDisableClientState (GL_COLOR_ARRAY);

		glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		glColor4f (1, 1, 1, 1);

		glDisable (GL_BLEND);
		glDepthMask (GL_TRUE);

		// update any particles that were drawn
		R_UpdateParticles ();
	}
}

