/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// contains common externs and defs for gl_*.c only stuff to avoid polluting glquake.h/quakedef.h
// and having to rebuild the entire project everytime anything is added/changed.  this is, in other
// words, what glquake.h SHOULD have been...


#include "quakedef.h"

#define FOG_RANGE 700.0f
#define FOG_ALPHA_MAX 0.9f
#define FOG_ALPHA_LAVA_CONST 0.75f
#define FOG_ALPHA_SLIME_CONST 0.666f
#define BUMPRESCALE 1.356382978723f

// GL implementation and hardware capabilities
typedef struct gl_caps_t
{
	int		GL_TextureUnits;
	float	GL_MaxAnisotropy;
	int		GL_MaxTextureSize;
	int		GL_OcclusionBits;
} gl_caps_s;

extern gl_caps_s GLCaps;

// useful stuff - only used during the render, it makes no sense at all to have it in
// a header file causing the entire project to be recompiled
void R_SetupGL (void);
void R_RotateForEntity (entity_t *e, qboolean blendLerp);
void GL_MakeTextureObjects (GLsizei n, GLuint *textures);
GLuint GL_MakeDisplayListObjects (GLsizei n);
void GL_BlendTest (void);
void GL_InitState (void);

// HUD model loading
int Mod_Load3DWidget (char *WidgetName);
int Mod_Load3DSkinnedWidget (char *WidgetName, int SkinNum, int framenum);

// for immediate drawing
void Mod_Draw3DSkinnedWidget (char *WidgetName, int SkinNum, int framenum);

// mainly for sky control
extern cvar_t r_farclip;
extern cvar_t r_skyspeed;


// maintain some info relating to the current GL state.  not really used outside of initialization.
// it was originally used for managing state changes, so no redundant ones would happen, but instead
// i just decided to write cleaner code...
typedef struct gl_state_s
{
	// utility stuff
	int				MaxHackTexW;
	int				MaxHackTexH;

	const char		*gl_vendor;
	const char		*gl_renderer;
	float			gl_version;
	const char		*gl_extensions;

	int				OrthoWidth;
	int				OrthoHeight;

	qboolean		OrthoSet;
	qboolean		NP2BigDirtyFib;
} gl_state_t;


// extension tracking
#define NO_EXTENSIONS			0
#define HAS_ANISOTROPIC			1
#define HAS_OCCLUSION			2
#define HAS_CVA					4
#define HAS_ARBVERTEXPROGRAMS	8
#define HAS_ARBFRAGMENTSHADERS	16

extern unsigned int GL_ExtensionBits;
extern gl_state_t	GL_State;

// texture flags
#define TEX_TEXTURE			0			// is a texture!!!
#define TEX_MIPMAP			1			// texture should be mipmapped
#define TEX_32BPP			2			// texture comes into the loader already at 32 bpp
#define TEX_BSP				4			// is a BSP texture
#define TEX_MDL				8			// is an MDL texture
#define TEX_ALPHAMASK		16			// is an alpha mask - alpha will be set to the avg of r/g/b, r/g/b will be 255
#define TEX_EXTERN			32			// is external
#define TEX_LIGHTMAP		64			// is a lightmap texture
#define TEX_LIQUID			128			// is a liquid texture
#define TEX_INCACHE			256			// texture struct came from the cache
#define TEX_HASLUMA			512			// texture has luma components
#define TEX_HASALPHA		1024		// texture has alpha components
#define TEX_MAKENORMAL		2048		// use to make a normal map
#define TEX_RESOURCE		4096		// texture loaded from a resource
#define TEX_NOGEN			8192		// don't generate or bind a texture object - it's assumed this is done outside of the loader
#define TEX_NOAGE			16384		// explicitly don't age the texture
#define TEX_LOADEDOK		32768		// texture was successfully loaded
#define TEX_NOCOMPRESS		65536

// texture loading entry points
gltexture_t *GL_LoadQuakeTexture (gltexture_t *glt, char *identifier, void *data, int width, int height, unsigned int flags);
int GL_UploadTextureToOpenGL (void *data, int width, int height, unsigned int flags);
qboolean GL_LoadBitmapResourceTexture (gltexture_t *glt, int ResourceID, unsigned int flags);

byte *GL_LoadPNG (gltexture_t *glt, char *basename);
byte *GL_LoadTGA (gltexture_t *glt, char *basename);
byte *GL_LoadBMP (gltexture_t *glt, char *basename);
byte *GL_LoadPCX (gltexture_t *glt, char *basename);

// texture cache
#define MAX_GL_TEXTURES MAX_MAP_TEXTURES * 4

void GL_DeleteAllTextures (void);
void GL_FlushTextureCache (void);
void GL_Compile_ARB_Programs (void);

typedef enum
{
	F_PROG_HEATHAZEMASK = 0,
	F_PROG_WARP,
	F_PROG_WATER_DISTORT,
	NUM_FRAGMENT_PROGRAM
} fr_progs;

typedef enum
{
    V_PROG_DISTORT = 0,
	NUM_VERTEX_PROGRAM
} vrt_progs;

extern GLuint fragment_programs[NUM_FRAGMENT_PROGRAM];
extern GLuint vertex_programs[NUM_VERTEX_PROGRAM];

// standard required textures
extern gltexture_t texturecache[MAX_GL_TEXTURES];
extern gltexture_t particletexture;
extern gltexture_t CrosshairTexture;
extern gltexture_t r_detailtexture;
extern gltexture_t r_caustictexture;
extern gltexture_t r_dst_texture;

// default colour based textures.  they're not really used (if at all) but are here in case
// they ever do come in handy
extern gltexture_t r_gamma;
extern gltexture_t r_white;
extern gltexture_t r_black;
extern gltexture_t r_red;
extern gltexture_t r_green;
extern gltexture_t r_blue;

// get the leaf an entity is in according to certain criteria.  the entity bounding box
// is used rather than just a single-point origin, because some entities may occupy more
// than 1 leaf, and we might want to pick one of them based on the viewleaf.
mleaf_t *R_LeafForEntity (vec3_t mins, vec3_t maxs);

// draw a particle sprite.  these are particles that live for 1 frame, and are refreshed every time.
void R_ParticleSprite (int color, float radius, vec3_t origin, pdrawstyle_t type);

// extra smoke effects
void R_SmokeSplash (vec3_t org);
void R_SmokeEffect (vec3_t org, int count, float InitialAlpha);

#define MAX_GL_STRIP_TRIS 128

// generic useful function type things
qboolean R_BoxContainsPoint (vec3_t p, vec3_t mins, vec3_t maxs, float fudge);
void R_DrawBBox (vec3_t mins, vec3_t maxs);
