/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// mathlib.c -- math primitives

#include <math.h>
#if _MSC_VER 
#include <time.h>
#include <xmmintrin.h>
#endif
#include "quakedef.h"

void Sys_Error (char *error, ...);

vec3_t vec3_origin = {0,0,0};

/*-----------------------------------------------------------------*/

float anglemod(float a)
{
	a = (360.0/65536) * ((int)(a*(65536/360.0)) & 65535);
	return a;
}

float angledelta (float a)
{
	a = anglemod(a);
	
	if (a > 180)
	{
		a -= 360;
	}
	return a;
}

/*-----------------------------------------------------------------*/

/*
==================
BOPS_Error

Split out like this for ASM to call.
==================
*/
void BOPS_Error (void)
{
	Sys_Error ("BoxOnPlaneSide:  Bad signbits");
}

/*
==================
BoxOnPlaneSide

Returns 1, 2, or 1 + 2
==================
*/
#if	!id386
int BoxOnPlaneSide (vec3_t emins, vec3_t emaxs, mplane_t *p)
{
	float	dist1, dist2;
	int		sides;

// general case
	switch (p->signbits)
	{
	case 0:
		dist1 = p->normal[0]*emaxs[0] + p->normal[1]*emaxs[1] + p->normal[2]*emaxs[2];
		dist2 = p->normal[0]*emins[0] + p->normal[1]*emins[1] + p->normal[2]*emins[2];
		break;
	case 1:
		dist1 = p->normal[0]*emins[0] + p->normal[1]*emaxs[1] + p->normal[2]*emaxs[2];
		dist2 = p->normal[0]*emaxs[0] + p->normal[1]*emins[1] + p->normal[2]*emins[2];
		break;
	case 2:
		dist1 = p->normal[0]*emaxs[0] + p->normal[1]*emins[1] + p->normal[2]*emaxs[2];
		dist2 = p->normal[0]*emins[0] + p->normal[1]*emaxs[1] + p->normal[2]*emins[2];
		break;
	case 3:
		dist1 = p->normal[0]*emins[0] + p->normal[1]*emins[1] + p->normal[2]*emaxs[2];
		dist2 = p->normal[0]*emaxs[0] + p->normal[1]*emaxs[1] + p->normal[2]*emins[2];
		break;
	case 4:
		dist1 = p->normal[0]*emaxs[0] + p->normal[1]*emaxs[1] + p->normal[2]*emins[2];
		dist2 = p->normal[0]*emins[0] + p->normal[1]*emins[1] + p->normal[2]*emaxs[2];
		break;
	case 5:
		dist1 = p->normal[0]*emins[0] + p->normal[1]*emaxs[1] + p->normal[2]*emins[2];
		dist2 = p->normal[0]*emaxs[0] + p->normal[1]*emins[1] + p->normal[2]*emaxs[2];
		break;
	case 6:
		dist1 = p->normal[0]*emaxs[0] + p->normal[1]*emins[1] + p->normal[2]*emins[2];
		dist2 = p->normal[0]*emins[0] + p->normal[1]*emaxs[1] + p->normal[2]*emaxs[2];
		break;
	case 7:
		dist1 = p->normal[0]*emins[0] + p->normal[1]*emins[1] + p->normal[2]*emins[2];
		dist2 = p->normal[0]*emaxs[0] + p->normal[1]*emaxs[1] + p->normal[2]*emaxs[2];
		break;
	default:
		dist1 = dist2 = 0;		// shut up compiler
		BOPS_Error ();
		break;
	}
	sides = 0;

	if (dist1 >= p->dist)
	{
		sides = 1;
	}

	if (dist2 < p->dist)
	{
		sides |= 2;
	}
	return sides;
}
#endif

void VectorAngles (const vec3_t forward, vec3_t angles)
{
	vec3_t temp;

	temp[0] = forward[0];
	temp[1] = forward[1];
	temp[2] = 0;
	angles[PITCH] = -atan2(forward[2], Length(temp)) / M_PI / 180.0f;
	angles[YAW] = atan2(forward[1], forward[0]) / M_PI / 180.0f;
	angles[ROLL] = 0;
}

void AngleVectors (vec3_t angles, vec3_t forward, vec3_t right, vec3_t up)
{
	float		angle;
	float		sr, sp, sy, cr, cp, cy;
	
	angle = angles[YAW] * (M_PI*2 / 360);
	sy = sin(angle);
	cy = cos(angle);
	angle = angles[PITCH] * (M_PI*2 / 360);
	sp = sin(angle);
	cp = cos(angle);
	angle = angles[ROLL] * (M_PI*2 / 360);
	sr = sin(angle);
	cr = cos(angle);

	forward[0] = cp*cy;
	forward[1] = cp*sy;
	forward[2] = -sp;
	right[0] = (-1*sr*sp*cy+-1*cr*-sy);
	right[1] = (-1*sr*sp*sy+-1*cr*cy);
	right[2] = -1*sr*cp;
	up[0] = (cr*sp*cy+-sr*-sy);
	up[1] = (cr*sp*sy+-sr*cy);
	up[2] = cr*cp;
}

void TurnVector (vec3_t out, const vec3_t forward, const vec3_t side, float angle)
{
	float scale_forward, scale_side;
	
	scale_forward = cos (DEG2RAD (angle));
	scale_side = sin (DEG2RAD (angle));
	
	out[0] = scale_forward * forward[0] + scale_side * side[0];
	out[1] = scale_forward * forward[1] + scale_side * side[1];
	out[2] = scale_forward * forward[2] + scale_side * side[2];
}

void VectorSubtract (vec3_t veca, vec3_t vecb, vec3_t out) 
{ 
#if _MSC_VER 
   // raynorpat: msvc sse optimization 
   __m128 xmm_veca, xmm_vecb, xmm_out; 

   xmm_veca = _mm_load_ss(&veca[0]); 
   xmm_vecb = _mm_load_ss(&vecb[0]); 
   xmm_out = _mm_sub_ss(xmm_veca,xmm_vecb); 
   _mm_store_ss(&out[0],xmm_out); 

   xmm_veca = _mm_load_ss(&veca[1]); 
   xmm_vecb = _mm_load_ss(&vecb[1]); 
   xmm_out = _mm_sub_ss(xmm_veca,xmm_vecb); 
   _mm_store_ss(&out[1],xmm_out); 

   xmm_veca = _mm_load_ss(&veca[2]); 
   xmm_vecb = _mm_load_ss(&vecb[2]); 
   xmm_out = _mm_sub_ss(xmm_veca,xmm_vecb); 
   _mm_store_ss(&out[2],xmm_out); 
#else 
   out[0] = veca[0]-vecb[0]; 
   out[1] = veca[1]-vecb[1]; 
   out[2] = veca[2]-vecb[2]; 
#endif 
} 

void VectorAdd (vec3_t veca, vec3_t vecb, vec3_t out) 
{ 
#if _MSC_VER 
   // raynorpat: msvc sse optimization 
   __m128 xmm_veca, xmm_vecb, xmm_out; 

   xmm_veca = _mm_load_ss(&veca[0]); 
   xmm_vecb = _mm_load_ss(&vecb[0]); 
   xmm_out = _mm_add_ss(xmm_veca,xmm_vecb); 
   _mm_store_ss(&out[0],xmm_out); 

   xmm_veca = _mm_load_ss(&veca[1]); 
   xmm_vecb = _mm_load_ss(&vecb[1]); 
   xmm_out = _mm_add_ss(xmm_veca,xmm_vecb); 
   _mm_store_ss(&out[1],xmm_out); 

   xmm_veca = _mm_load_ss(&veca[2]); 
   xmm_vecb = _mm_load_ss(&vecb[2]); 
   xmm_out = _mm_add_ss(xmm_veca,xmm_vecb); 
   _mm_store_ss(&out[2],xmm_out); 
#else 
   out[0] = veca[0]+vecb[0]; 
   out[1] = veca[1]+vecb[1]; 
   out[2] = veca[2]+vecb[2]; 
#endif 
} 

void VectorCopy (vec3_t in, vec3_t out) 
{ 
#if _MSC_VER 
   // raynorpat: msvc sse optimization 
   __m128 xmm_in; 

   xmm_in = _mm_load_ss(&in[0]); 
   _mm_store_ss(&out[0],xmm_in); 

   xmm_in = _mm_load_ss(&in[1]); 
   _mm_store_ss(&out[1],xmm_in); 

   xmm_in = _mm_load_ss(&in[2]); 
   _mm_store_ss(&out[2],xmm_in); 
#else 
   out[0] = in[0]; 
   out[1] = in[1]; 
   out[2] = in[2]; 
#endif 
} 

void VectorScale (vec3_t in, vec_t scale, vec3_t out) 
{ 
#if _MSC_VER 
   // raynorpat: msvc sse optimization 
   __m128 xmm_in, xmm_scale, xmm_out; 

   xmm_in = _mm_load_ss(&in[0]); 
   xmm_scale = _mm_load_ss(&scale); 
   xmm_out = _mm_mul_ss(xmm_in,xmm_scale); 
   _mm_store_ss(&out[0],xmm_out); 

   xmm_in = _mm_load_ss(&in[1]); 
   xmm_scale = _mm_load_ss(&scale); 
   xmm_out = _mm_mul_ss(xmm_in,xmm_scale); 
   _mm_store_ss(&out[1],xmm_out); 

   xmm_in = _mm_load_ss(&in[2]); 
   xmm_scale = _mm_load_ss(&scale); 
   xmm_out = _mm_mul_ss(xmm_in,xmm_scale); 
   _mm_store_ss(&out[2],xmm_out); 
#else 
   out[0] = in[0]*scale; 
   out[1] = in[1]*scale; 
   out[2] = in[2]*scale; 
#endif 
} 

float Length(vec3_t inout)
{
	int		i;
	float	length;
	
	length = 0;
	
	for (i=0 ; i< 3 ; i++)
	{
		length += inout[i]*inout[i];
	}
	length = sqrt((double)length);		// FIXME
	
	return length;
}

float Distance(const vec3_t v, const vec3_t v2)
{
	int		i;
	float	length, d;
	
	length = 0;
	
	for (i=0 ; i< 3 ; i++)
	{
		d = v[i] - v2[i];
		length += d*d;
	}
	length = sqrt ((double)length);		// FIXME
	
	return length;
}

float Normalize (vec3_t inout)
{
	float	length, ilength;
	
	length = sqrt ((double)inout[0]*inout[0] + inout[1]*inout[1] + inout[2]*inout[2]);
	
	if (!length)
	{
		VectorClear (inout);
		return 0;
	}	
	ilength = 1.0/length;
	inout[0] = inout[0]*ilength;
	inout[1] = inout[1]*ilength;
	inout[2] = inout[2]*ilength;
	
	return length;
}

// fast version stolen from q3a
void NormalizeFast (vec3_t inout)
{
	float ilength;

	ilength = sqrt ((double)inout[0]*inout[0] + inout[1]*inout[1] + inout[2]*inout[2]);

	inout[0] *= ilength;
	inout[1] *= ilength;
	inout[2] *= ilength;
}

void VectorInverse (vec3_t v)
{
	v[0] = -v[0];
	v[1] = -v[1];
	v[2] = -v[2];
}

/*
===================
FloorDivMod

Returns mathematically correct (floor-based) quotient and remainder for
numer and denom, both of which should contain no fractional part. The
quotient must fit in 32 bits.
====================
*/

void FloorDivMod (double numer, double denom, int *quotient, int *rem)
{
	int		q, r;
	double	x;

	if (denom <= 0.0)
	{
		Sys_Error ("FloorDivMod: bad denominator %d\n", denom);
	}

	if (numer >= 0.0)
	{

		x = floor(numer / denom);
		q = (int)x;
		r = (int)floor(numer - (x * denom));
	}
	else
	{
	//
	// perform operations with positive values, and fix mod to make floor-based
	//
		x = floor(-numer / denom);
		q = -(int)x;
		r = (int)floor(-numer - (x * denom));

		if (r != 0)
		{
			q--;
			r = (int)denom - r;
		}
	}
	*quotient = q;
	*rem = r;
}


/*
===================
GreatestCommonDivisor
====================
*/
int GreatestCommonDivisor (int i1, int i2)
{
	if (i1 > i2)
	{
		if (i2 == 0)
		{
			return (i1);
		}
		return GreatestCommonDivisor (i2, i1 % i2);
	}
	else
	{
		if (i1 == 0)
		{
			return (i2);
		}
		return GreatestCommonDivisor (i1, i2 % i1);
	}
}

/*
===================
Invert24To16

Inverts an 8.24 value to a 16.16 value
====================
*/
#if	!id386
fixed16_t Invert24To16(fixed16_t val)
{
	if (val < 256) 
	{
		return (0xFFFFFFFF);
	}
	return (fixed16_t) (((double)0x10000 * (double)0x1000000 / (double)val) + 0.5);
}
#endif


/*
========================================================================================================================

									OPTIMIZED RANDOM NUMBER GENERATOR

	Contains a list of precalculated random numbers in an array large enough to ensure that continuous use doesn't
	show up obvious repititions in the pattern.  The numbers are calculated at load time and during other events
	when gameplay pauses, to add further variety and robustness.  Generating a random number is just a matter of
	selecting the next available number from the array.  If we get to the end of the array, we wrap around to
	the start again.

========================================================================================================================
*/

#define Q_RANDTABLE_SIZE 8192

int		Q_randtable[Q_RANDTABLE_SIZE];
int		Q_randindex;
float	Q_sqrttable[262144];

void Q_randinit (void)
{
	int i;

	// seed with the time
	srand ((unsigned) time (NULL));

	// build a table of random numbers for fast lookups
	for (i = 0; i < Q_RANDTABLE_SIZE; i++)
	{
		Q_randtable[i] = rand ();
	}
	Q_randindex = 0;

	for (i = 0; i < 262144; i++)
	{
		Q_sqrttable[i] = sqrt (i);
	}
}

// stolen from q3a
float Q_rsqrt (float num)
{
	long		i;
	float		x2, y;
	const float threehalfs = 1.5F;
	
	x2 = num * 0.5F;
	y  = num;
	i  = *(long *) &y;							// evil floating point bit level hacking
	i  = 0x5f3759df - (i >> 1);					// what the fuck?
	y  = *(float *) &i;
	y  = y * (threehalfs - (x2 * y * y));		// 1st iteration
	
	return y;
}

float Q_sqrt (float num)
{
	int i;

	if (num < 4)
	{
		return Q_rsqrt (num);
	}

	if (num < 262144)
	{
		return Q_sqrttable[(int) num];
	}

	for (i = 512; ; i++)
	{
		if (i * i > num) 
		{
			return i;
		}
	}
	return Q_rsqrt (num);
}

int Q_rand (void)
{
	(Q_randindex == Q_RANDTABLE_SIZE) ? Q_randindex = 0 : Q_randindex++;

	return Q_randtable[Q_randindex];
}

int Q_log2(int val)
{
	int answer = 0;
	
	while (val>>=1)
	{
		answer++;
	}
	return answer;
}