/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// snd_mem.c: sound caching

#include "quakedef.h"

int			cache_full_cycle;

byte *S_Alloc (int size);

/*
================
ResampleSfx
================
*/
void ResampleSfx (sfx_t *sfx, int inrate, int inwidth, byte *data)
{
	int			outcount;
	int			srcsample;
	float		stepscale;
	int			i;
	int			sample, samplefrac, fracstep;
	
	if (!sfx->loaded) return;

	stepscale = (float)inrate / shm->speed;	// this is usually 0.5, 1, or 2

	outcount = sfx->length / stepscale;
	sfx->length = outcount;

	if (sfx->loopstart != -1)
	{
		sfx->loopstart = sfx->loopstart / stepscale;
	}
	sfx->speed = shm->speed;

	if (loadas8bit.value)
	{
		sfx->width = 1;
	}
	else
	{
		sfx->width = inwidth;
	}
	sfx->stereo = 0;

	// resample / decimate to the current source rate
	if (stepscale == 1 && inwidth == 1 && sfx->width == 1)
	{
		// fast special case
		for (i=0 ; i<outcount ; i++)
		{
			((signed char *)sfx->data)[i] = (int)( (unsigned char)(data[i]) - 128);
		}
	}
	else
	{
		// general case
		samplefrac = 0;
		fracstep = stepscale*256;

		for (i=0 ; i<outcount ; i++)
		{
			srcsample = samplefrac >> 8;
			samplefrac += fracstep;

			if (inwidth == 2)
			{
				sample = LittleShort ( ((short *)data)[srcsample] );
			}
			else
			{
				sample = (int)( (unsigned char)(data[srcsample]) - 128) << 8;
			}

			if (sfx->width == 2)
			{
				((short *)sfx->data)[i] = sample;
			}
			else
			{
				((signed char *)sfx->data)[i] = sample >> 8;
			}
		}
	}
}

//=============================================================================

/*
===================
S_ClearAll
===================
*/
void S_ClearAll (void)
{
	int			i;
	sfx_t		*sfx;

	// reload all sounds at map start
	for (i = 0 ,sfx = known_sfx ; i < num_sfx ; i++, sfx++)
	{
		if (sfx->loaded && sfx->name[0])
		{
			sfx->loaded = false;
		}
	}
}

/*
==============
S_LoadSound
==============
*/
sfx_t *S_LoadSound (sfx_t *s)
{
    char		namebuffer[256];
	wavinfo_t	info;
	int			len;
	float		stepscale;
	byte		*data;

	if (s->loaded)
	{
		return s;
	}

	// load it in and make sure it fits !!!
	strlcpy_s (namebuffer, "sound/");
	strlcat_s (namebuffer, s->name);

	// load it in the heap
	if (!(data = COM_LoadTempFile(namebuffer)))
	{
		Con_Printf ("Couldn't load %s\n", namebuffer);
		return NULL;
	}
	info = GetWavinfo (s->name, data, com_filesize);

	if (info.channels != 1)
	{
		Con_Printf ("%s is a stereo sample\n",s->name);
		return NULL;
	}
	stepscale = (float)info.rate / shm->speed;	
	len = info.samples / stepscale;
	len = len * info.width * info.channels;

	if (!info.samples)
	{
		len = 1;
	}
	s->data = Heap_ZAlloc (len + sizeof (sfx_t));

	if (!s->data) 
	{
		return NULL;
	}
	s->loaded = true;
	s->length = info.samples;
	s->loopstart = info.loopstart;
	s->speed = info.rate;
	s->width = info.width;
	s->stereo = info.channels;

	ResampleSfx (s, s->speed, s->width, data + info.dataofs);

	return s;
}



/*
===============================================================================

WAV loading

===============================================================================
*/


byte	*data_p;
byte 	*iff_end;
byte 	*last_chunk;
byte 	*iff_data;
char	*iff_name;
int 	iff_chunk_len;


// GetLittleShort is used by DevIL so i renamed it...
short SNDGetLittleShort(void)
{
	short val = 0;
	val = *data_p;
	val = val + (*(data_p + 1) << 8);
	data_p += 2;
	return val;
}

int SNDGetLittleLong(void)
{
	int val = 0;
	val = *data_p;
	val = val + (*(data_p + 1) << 8);
	val = val + (*(data_p + 2) << 16);
	val = val + (*(data_p + 3) << 24);
	data_p += 4;
	return val;
}

void FindNextChunk(const char *name)
{
	while (1)
	{
		data_p = last_chunk;

		if (data_p >= iff_end)
		{	
			// didn't find the chunk
			data_p = NULL;
			return;
		}		
		data_p += 4;
		iff_chunk_len = SNDGetLittleLong();

		if (iff_chunk_len < 0)
		{
			data_p = NULL;
			return;
		}
		data_p -= 8;

		if (iff_end - data_p < 8 + iff_chunk_len) 
		{
			Con_Printf("Corrupt sound file '%s'.\n", iff_name);
			data_p = NULL;
			return;
		}
		last_chunk = data_p + 8 + ( (iff_chunk_len + 1) & ~1 );

		if (!memcmp (data_p, name, 4)) return;
	}
}

void FindChunk(const char *name)
{
	last_chunk = iff_data;
	FindNextChunk (name);
}

void DumpChunks(void)
{
	char	str[5];
	
	str[4] = 0;
	data_p=iff_data;

	do
	{
		memcpy (str, data_p, 4);
		data_p += 4;
		iff_chunk_len = SNDGetLittleLong();
		Con_Printf ("0x%x : %s (%d)\n", (int)(data_p - 4), str, iff_chunk_len);
		data_p += (iff_chunk_len + 1) & ~1;
	} while (data_p < iff_end);
}

/*
============
GetWavinfo
============
*/
wavinfo_t GetWavinfo (char *name, byte *wav, int wavlength)
{
	wavinfo_t	info;
	int			i;
	int			format;
	int			samples;

	memset (&info, 0, sizeof(info));

	if (!wav || !wavlength)
	{
		return info;
	}
	iff_name = name;
	iff_data = wav;
	iff_end = wav + wavlength;

	// find "RIFF" chunk
	FindChunk("RIFF");

	if (!(data_p && !memcmp (data_p + 8, "WAVE", 4)))
	{
		Con_Printf("Missing RIFF/WAVE chunks\n");
		return info;
	}

	// get "fmt " chunk
	iff_data = data_p + 12;

	if (developer.value)
	{
		DumpChunks ();
	}
	FindChunk("fmt ");

	if (!data_p)
	{
		Con_Printf("Missing fmt chunk\n");
		return info;
	}
	data_p += 8;
	format = SNDGetLittleShort();

	if (format != 1)
	{
		Con_Printf("Microsoft PCM format only\n");
		return info;
	}
	info.channels = SNDGetLittleShort();
	info.rate = SNDGetLittleLong();
	data_p += 4+2;
	info.width = SNDGetLittleShort() / 8;

	// get cue chunk
	FindChunk("cue ");

	if (data_p)
	{
		data_p += 32;
		info.loopstart = SNDGetLittleLong();

		// if the next chunk is a LIST chunk, look for a cue length marker
		FindNextChunk ("LIST");

		if (data_p)
		{
			if (!memcmp (data_p + 28, "mark", 4))
			{	
				// this is not a proper parse, but it works with cooledit...
				data_p += 24;
				i = SNDGetLittleLong ();	// samples in loop
				info.samples = info.loopstart + i;
			}
		}
	}
	else
	{
		info.loopstart = -1;
	}
	// find data chunk
	FindChunk("data");

	if (!data_p)
	{
		Con_Printf("Missing data chunk\n");
		return info;
	}
	data_p += 4;
	samples = SNDGetLittleLong () / info.width;

	if (info.samples)
	{
		if (samples < info.samples)
		{
			Sys_Error ("Sound %s has a bad loop length", name);
		}
	}
	else
	{
		info.samples = samples;
	}
	info.dataofs = data_p - wav;
	
	return info;
}

